"""
WHAD provides various classes to interact with WHAD-enabled hardware:

- :py:class:`whad.device.device.Device`
- :py:class:`whad.device.device.VirtualDevice`

Class :class:`whad.device.device.Device` is the default class that allows WHAD devices
enumeration and access. It is the main class to use to open any device, through
its :method:`whad.device.Device.create` method as shown below:

.. code-block:: python

    from whad.device import Device

    dev = Device.create("uart0")

The :py:class:`whad.device.device.VirtualDevice` shall not be directly used. This class
is used to add support for incompatible WHAD devices like the *Ubertooth*
or the *ApiMote* and acts as an adaptation layer between the underlying WHAD
protocol and the specific protocol used by the target hardware.

.. important::

    The :py:class:`whad.device.device.WhadDevice` class that is still defined in WHAD (and used
    in some old example scripts or documentation) is an alias for the new
    :py:class:`whad.device.Device` class, and is meant to be deprecated in the future.
    This old class has been renamed to ``Device`` for clarity, and the same happened
    with the old default connector class :py:class:`whad.device.connector.WhadConnector`
    that has been renamed to :py:class:`whad.device.connector.Connector`.

    These old classes will be marked as *deprecated* in a future release, with a
    specific EOL date announced. A warning message will be issued in case one of
    these classes is used in a script or a tool to give time to users to migrate
    to the new ones (renaming classes is enough to switch to the new implementation,
    APIs stay the same).

Default device classes
----------------------

.. autoclass:: whad.device.device.Device
    :members:

.. autoclass:: whad.device.device.VirtualDevice
    :show-inheritance:
    :members:

Old device classes to be deprecated in the future
-------------------------------------------------

.. autoclass:: whad.device.device.WhadDevice
    :show-inheritance:

.. autoclass:: whad.device.device.WhadVirtualDevice
    :show-inheritance:

"""
import re
import logging
import contextlib
from time import time
from typing import Generator, Callable, Union, Type, Optional
from threading import Thread, Lock
from queue import Queue, Empty

from whad.helpers import message_filter
from whad.exceptions import (
    WhadDeviceNotReady, WhadDeviceError, WhadDeviceTimeout, WhadDeviceDisconnected,
    WhadDeviceNotFound,
)
from whad.hub import ProtocolHub
from whad.hub.message import HubMessage
from whad.hub.generic.cmdresult import CommandResult
from whad.hub.discovery import InfoQueryResp, DomainInfoQueryResp, DeviceReady
from whad.hub.discovery import DeviceType

from .info import DeviceInfo

logger = logging.getLogger(__name__)

class DeviceEvt:
    """Device event.
    """
    def __init__(self, device: Optional['Device'] = None):
        self.__device = device

    @property
    def device(self) -> Optional['Device']:
        """Related device"""
        return self.__device

    def __repr__(self) -> str:
        """Printable representation of this event
        """
        iface = self.device.interface if self.device is not None else 'undefined'
        return f"DeviceEvt(iface='{iface}')"

class Disconnected(DeviceEvt):
    """Interface has disconnected.
    """
    def __repr__(self) -> str:
        """Printable representation of this event.
        """
        iface = self.device.interface if self.device is not None else 'undefined'
        return f"Disconnected(device='{iface}')"

class MessageReceived(DeviceEvt):
    """Message has been received.
    """
    def __init__(self, device = None, message = None):
        super().__init__(device)
        self.__message = message

    @property
    def message(self) -> Optional[HubMessage]:
        """Received message."""
        return self.__message

    def __repr__(self) -> str:
        """Printable representation of this event.
        """
        iface = self.device.interface if self.device is not None else 'undefined'
        return f"MessageReceived(device='{iface}')"

class DevInThread(Thread):
    """Internal thread processing data sent by the connector to the
    hardware interface.
    """

    def __init__(self, device = None):
        super().__init__(daemon=True)
        self.__iface = device
        self.__canceled = False

    def cancel(self):
        """Cancel thread
        """
        self.__canceled = True

    def serialize(self, message) -> bytes:
        """Serialize a WHAD message.
        """
        # Serialize protobuf message
        raw_msg = message.serialize()

        # Define header
        header = [
            0xAC, 0xBE,
            len(raw_msg) & 0xff,
            (len(raw_msg) >> 8) & 0xff
        ]

        # Build the final payload
        return bytes(header) + raw_msg

    def run(self):
        """Out thread main task.
        """
        while not self.__canceled:
            # Read data from device (may block)
            try:
                # Wait for a message to send to interface (blocking)
                logger.debug("[%s][in_thread] waiting for message to send", self.__iface.interface)
                with self.__iface.get_pending_message(timeout=1.0) as message:
                    logger.debug("[%s][in_thread] sending message %s",
                                 self.__iface.interface, message)

                    # Serialize message and send it.
                    payload = self.serialize(message)

                    # Send serialized message to interface
                    logger.debug("[%s][in_thread] acquiring lock on interface ...",
                                 self.__iface.interface)

                    self.__iface.lock()

                    logger.debug("[%s][in_thread] sending payload %s ...",
                                 self.__iface.interface, payload)
                    self.__iface.write(payload)
                    logger.debug("[%s][in_thread] releasing lock ...", self.__iface.interface)
                    self.__iface.unlock()

                    # Notify message has correctly been sent, from a dedicated
                    # thread.
                    if message.has_callback():
                        Thread(target=message.sent).start()
            except Empty:
                pass
            except WhadDeviceNotReady:
                if message.has_callback():
                    Thread(target=message.error, args=[1]).start()
                break
            except WhadDeviceDisconnected:
                if message.has_callback():
                    Thread(target=message.error, args=[2]).start()
                break

class DevOutThread(Thread):
    """Internal thread processing data sent by the hardware interface
    to the device object.
    """
    def __init__(self, device = None):
        super().__init__(daemon=True)
        self.__iface = device
        self.__canceled = False


        # Data processing
        self.__data = bytearray()

    def cancel(self):
        """Cancel thread
        """
        self.__canceled = True

    def ingest(self, data: bytes):
        """Ingest incoming bytes.
        """
        self.__data.extend(data)
        while len(self.__data) > 2:
            # Is the magic correct ?
            if self.__data[0] == 0xAC and self.__data[1] == 0xBE:
                # Have we received a complete message ?
                if len(self.__data) > 4:
                    msg_size = self.__data[2] | (self.__data[3] << 8)
                    if len(self.__data) >= (msg_size+4):
                        raw_message = self.__data[4:4+msg_size]

                        # Parse received message with our Protocol Hub
                        msg = self.__iface.hub.parse(bytes(raw_message))

                        # Forward message if successfully parsed
                        if msg is not None:
                            self.__iface.put_message(msg)

                        # Chomp
                        self.__data = self.__data[msg_size + 4:]
                    else:
                        break
                else:
                    break
            else:
                # Nope, that's not a header
                while len(self.__data) >= 2:
                    if (self.__data[0] != 0xAC) or (self.__data[1] != 0xBE):
                        self.__data = self.__data[1:]
                    else:
                        break

    def run(self):
        """Out thread main task.
        """
        while not self.__canceled:
            # Read data from device (may block)
            try:
                data = self.__iface.read()
                if data is not None:
                    self.ingest(data)
            except WhadDeviceNotReady:
                break
            except WhadDeviceDisconnected:
                # Device has disconnected, notify interface by injecting a
                # `Disconnected` event message in its output message queue.
                # This event message will be filtered out by the Interface class
                # when called from our connector thread, and it will exit
                # gracefully
                logger.debug("[iface][%s] Device disconnected, sending event to connector.",
                             self.__iface.interface)
                self.__iface.put_message(Disconnected(self.__iface))
                return



class Device:
    """WHAD hardware interface
    """

    # Should be lowercase.
    INTERFACE_NAME = None

    @staticmethod
    def get_all_ifaces():
        """Load all known interfaces."""
        # Base transport
        from .uart import Uart
        from .hci import Hci
        from .tcp import TcpSocket
        from .unix import UnixSocket
        # Virtual transport
        from .apimote import Apimote
        from .pcap import Pcap
        from .rfstorm import RfStorm
        from .rzusbstick import RzUsbStick
        from .ubertooth import Ubertooth
        from .yard import YardStickOne
        return [
            Uart,
            Hci,
            TcpSocket,
            UnixSocket,
            Pcap,
            Apimote,
            RfStorm,
            RzUsbStick,
            Ubertooth,
            YardStickOne
        ]

    @classmethod
    def _get_sub_classes(cls):
        """
        Helper allowing to get every subclass of Device.
        """
        # List every available device class
        device_classes = set()
        for device_class in cls.__subclasses__():
            if device_class.__name__ in ("WhadVirtualDevice", "VirtualDevice"):
                for virtual_device_class in device_class.__subclasses__():
                    device_classes.add(virtual_device_class)
            else:
                device_classes.add(device_class)
        return device_classes

    @classmethod
    def create_inst(cls, interface_string) -> Type['Device']:
        """
        Helper allowing to get a device according to the interface string provided.

        To make it work, every device class must implement:
            - a class attribute INTERFACE_NAME, matching the interface name
            - a class method list, returning the available devices
            - a property identifier, allowing to identify the device in a unique way

        This method should NOT be used outside of this class. Use Device.create instead.
        """
        if cls.INTERFACE_NAME is None:
            raise WhadDeviceNotFound()

        logger.debug("Creating an instance of %s", interface_string)
        # Parses interface string, raise execption if format is incorrect
        pattern = r'^' + cls.INTERFACE_NAME + r'(:(?P<identifier>.*))?(?P<index>\d+)?$'
        result = re.match(pattern, interface_string)
        if result is None:
            raise WhadDeviceNotFound()

        # Extract interface identifier, if specified
        iface = result.groupdict()
        if iface['index'] is not None:
            iface_id = int(iface['index'])
        elif iface['identifier'] is not None:
            iface_id = iface['identifier']
        else:
            iface_id = None

        # Retrieve the list of available devices
        # and build a lookup dict
        interfaces = {}
        ifaces = cls.list()
        if isinstance(ifaces, list):
            for index, dev in enumerate(ifaces):
                interfaces[index] = dev
                interfaces[dev.identifier] = dev
        elif isinstance(ifaces, dict):
            for dev_id, dev in ifaces.items():
                interfaces[dev_id] = dev
                interfaces[dev.identifier] = dev
        else:
            interfaces = None

        # Some child classes may return None, in this case we need to use the
        # `check_interface()` to find the specified interface.
        if interfaces is None and iface_id is not None and cls.check_interface(iface_id):
            # Found, return an instance of this interface
            return cls(iface_id)

        if interfaces is not None and iface_id in interfaces:
            return interfaces[iface_id]

        if interfaces is not None and iface_id is None and 0 in interfaces and interfaces[0] is not None:
            return interfaces[0]

        # Return interface or raise an exception
        raise WhadDeviceNotFound()


    @classmethod
    def create(cls, interface_string):
        '''
        Create a specific device according to the provided interface string,
        formed as follows:

        <device_type>[device_index][:device_identifier]

        Examples:
            - `uart` or `uart0`: defines the first compatible UART device available
            - `uart1`: defines the second compatible UART device available
            - `uart:/dev/ttyACMO`: defines a compatible UART device identified
              by `/dev/tty/ACMO`
            - `ubertooth` or `ubertooth0`: defines the first available Ubertooth device
            - `ubertooth:11223344556677881122334455667788`: defines a Ubertooth
              device with serial number *11223344556677881122334455667788*
        '''
        # First try to find the corresponding device by calling find()
        device = Device.find(interface_string)
        if device is not None:
            return device

        # If not found, loop over all known interfaces
        device_classes = cls.get_all_ifaces()
        device = None
        for device_class in device_classes:
            logger.debug("trying class %s", device_class)
            try:
                device = device_class.create_inst(interface_string)
                return device
            except WhadDeviceNotFound:
                continue

        raise WhadDeviceNotFound

    @staticmethod
    def find(interface: str) -> Type['Device']:
        """Find a device based on its interface string with lazy loading of
        available interface implementations.

        :param interface: Interface name following WHAD's standard interface pattern
        :type  interface: str
        :return: Found interface object
        :rtype: Device
        :raise: WhadDeviceNotFound
        """
        # Based on selected transport, load the corresponding interface class.
        dev_info = re.match('^([^0-9:]+)([0-9]|:).*$', interface)
        if dev_info is not None:
            transport = dev_info.group(1).lower()
            # Basic interfaces using native transport
            if transport == 'uart':
                from .uart import Uart
                return Uart.create_inst(interface)
            elif transport == 'hci':
                from .hci import Hci
                return Hci.create_inst(interface)
            elif transport == 'tcp':
                from .tcp import TcpSocket
                return TcpSocket.create_inst(interface)
            elif transport == 'unix':
                from .unix import UnixSocket
                return UnixSocket.create_inst(interface)

            # Virtual interfaces using emulated transport
            elif transport == 'apimote':
                from .apimote import Apimote
                return Apimote.create_inst(interface)
            elif transport == 'pcap':
                from .pcap import Pcap
                return Pcap.create_inst(interface)
            elif transport == 'rfstorm':
                from .rfstorm import RfStorm
                return RfStorm.create_inst(interface)
            elif transport == 'rzusbstick':
                from .rzusbstick import RzUsbStick
                return RzUsbStick.create_inst(interface)
            elif transport == 'ubertooth':
                from .ubertooth import Ubertooth
                return Ubertooth.create_inst(interface)
            elif transport == 'yardstickone':
                from .yard import YardStickOne
                return YardStickOne.create_inst(interface)

            # No other known transport
            raise WhadDeviceNotFound


    @classmethod
    def list(cls) -> Union[list,dict]:
        '''
        Returns every available compatible devices.
        '''
        device_classes = cls.get_all_ifaces()

        available_devices = []
        for device_class in device_classes:
            # Reset device index before enumerating.
            device_class.reset_dev_index()

            # Enumerating
            device_class_list = device_class.list()
            if device_class_list is not None:
                if isinstance(device_class_list, list):
                    for device in device_class_list:
                        available_devices.append(device)
                elif isinstance(device_class_list, dict):
                    for _, device in device_class_list.items():
                        available_devices.append(device)
        return available_devices

    @classmethod
    def check_interface(cls, interface):
        '''
        Checks dynamically if the device can be instantiated.
        '''
        logger.debug("default: checking interface %s fails.", interface)
        return False

    @property
    def interface(self):
        '''
        Returns the current interface of the device.
        '''
        # If class has interface name, return the interface alias
        if hasattr(self.__class__,"INTERFACE_NAME"):
            if self.__class__.INTERFACE_NAME is not None:
                return self.__class__.INTERFACE_NAME + str(self.index)

        # Interface is unknown
        return "unknown"

    @property
    def type(self):
        '''
        Returns the name of the class linked to the current device.
        '''
        return self.__class__.__name__

    def __init__(self, index: Optional[int] = None):
        """Initialize an interface
        """
        # Interface state
        self.__info = None
        self.__opened = False
        self.__discovered = False

        # Generate device index if not provided
        if index is None:
            self.inc_dev_index()
            self.__index = self.__class__.CURRENT_DEVICE_INDEX
        elif isinstance(index, int):
            # Used by HCI devices to force index to match system names
            self.__index = index
        else:
            raise WhadDeviceNotFound()

        # IO Threads
        self.__iface_in = None
        self.__iface_out = None

        # Queue holding messages from connector, waiting to be sent to
        # the interface.
        self.__in_messages = Queue()

        # Queue holding messages from interface, waiting to be sent to
        # an attached connector
        self.__out_messages = Queue()

        # Connector bound to this device
        self.__connector = None

        # Interface lock
        self.__lock = Lock()
        # Connector lock
        self.__msg_filter: Callable[..., bool] = None

        # Protocol hub
        self.__hub = ProtocolHub()

        # Communication timeout
        self.__timeout = 5.0

    @contextlib.contextmanager
    def get_pending_message(self, timeout: float = None) -> Generator[HubMessage, None, None]:
        """Get message waiting to be sent to the interface.
        """
        try:
            yield self.__in_messages.get(timeout=timeout)
        except Empty as err:
            raise err

        # Mark task done
        self.__in_messages.task_done()

    @property
    def connector(self):
        """Connector bound to the interface
        """
        return self.__connector

    @property
    def hub(self):
        """Retrieve the device protocol hub (parser/factory)
        """
        return self.__hub

    @property
    def index(self) -> int:
        """Get the interface index

        :return: Interface index
        :rtype: int
        """
        return self.__index

    @property
    def device_id(self) -> Optional[str]:
        """Return device ID
        """
        if self.__info is not None:
            return self.__info.device_id

        # No info yet.
        return None

    @property
    def info(self) -> Optional[DeviceInfo]:
        """Get device info object

        :return: Device information object
        :rtype: DeviceInfo
        """
        return self.__info

    @property
    def opened(self) -> bool:
        """Device is open ?
        """
        return self.is_open()

    def is_open(self) -> bool:
        """Determine if interface is opened.
        """
        return self.__opened

    @classmethod
    def reset_dev_index(cls) -> None:
        """Reset the device index of the specified class.

        :param cls: Device class
        :type cls: Device
        """
        # Remove 'CURRENT_DEVICE_INDEX' attribute if set.
        if hasattr(cls, 'CURRENT_DEVICE_INDEX'):
            delattr(cls, 'CURRENT_DEVICE_INDEX')

    @classmethod
    def inc_dev_index(cls):
        """Inject and maintain device index.
        """
        if hasattr(cls, 'CURRENT_DEVICE_INDEX'):
            cls.CURRENT_DEVICE_INDEX += 1
        else:
            cls.CURRENT_DEVICE_INDEX = 0

    def set_connector(self, connector):
        """Set interface connector.
        """
        self.__connector = connector

    def lock(self):
        """Lock interface for read/write operation.
        """
        self.__lock.acquire()
        logger.debug("Lock acquired !")

    def unlock(self):
        """Unlock interface for read/write operation.
        """
        logger.debug("Releasing lock ...")
        self.__lock.release()

    def __start_io_threads(self):
        """Start background IO threads
        """
        self.__iface_in = DevInThread(self)
        self.__iface_in.start()
        self.__iface_out= DevOutThread(self)
        self.__iface_out.start()

    def __stop_io_threads(self):
        """Stop background IO threads
        """
        if self.__iface_in is not None:
            self.__iface_in.cancel()
        if self.__iface_out is not None:
            self.__iface_out.cancel()

    ##
    # Device specific methods
    ##

    def open(self):
        """Handle device open
        """
        # Create interface I/O threads and start them.
        self.__start_io_threads()

        # Ask interface for a reset
        try:
            logger.info("resetting interface (if possible)")
            self.__opened = True
            self.reset()
        except Empty as err:
            # Device is unresponsive, shutdown IO threads
            self.__stop_io_threads()

            raise WhadDeviceNotReady() from err

    def read(self) -> bytes:
        """Read bytes from interface (blocking).
        """
        return b''

    def write(self, payload: bytes) -> int:
        """Write payload to interface.
        """
        return len(payload)

    def close(self):
        """Close device
        """
        logger.info("closing WHAD interface")

        # Cancel I/O thread if required
        self.__stop_io_threads()
        self.__opened = False

        # Notify connector that device has closed
        if self.__connector is not None:
            logger.debug("Send disconnection event to connector %s", self.__connector)
            self.__connector.send_event(Disconnected(self))
            logger.debug("Disconnection event sent !")

    def change_transport_speed(self, speed):
        """Set device transport speed.

        Optional.
        """

    ##
    # Message processing
    ##

    def busy(self) -> bool:
        """Check if the interface is busy.

        We consider an interface as busy if there is at least one message in its
        output messages or in its input messages.
        """
        return not (self.__out_messages.empty() and self.__in_messages.empty())

    def set_queue_filter(self, keep: Callable[..., bool] = None):
        """Set message queue filter.
        """
        self.__msg_filter = keep

    def put_message(self, message: Union[HubMessage, DeviceEvt]):
        """Process incoming message.
        """
        # If no connector is attached to the interface, redirect to a dedicated
        # message queue. Same if the message is an interface event (this type of
        # messages MUST be handled by the interface itself).
        logger.debug("[%s] putting message %s", self.interface, message)
        if self.__connector is None:
            logger.debug("[%s] connector is None, sending to pending messages", self.interface)
            self.__out_messages.put(message)

        # If a connector is attached to the interface but a message filter
        # is set, redirect matching messages to a dedicated message queue
        # and notify the connector about the other messages, except if we
        # receive critical events from interface.
        elif isinstance(message, DeviceEvt):
            logger.debug("Sending event %s to connector %s", message, self.connector)
            self.connector.send_event(message)
        elif isinstance(message, HubMessage):
            # If a filter is set and message does not match, save it in our
            # pending messages queue.
            if self.__msg_filter is not None and self.__msg_filter(message):
                self.__out_messages.put(message)
            else:
                logger.debug("[%s] forwarding message to connector %s",
                             self.interface, self.connector)

                # Otherwise, wrap hub message into a `MessageReceived` event
                self.connector.send_event(MessageReceived(self, message))
        else:
            # Unknown message type, log it.
            logger.debug("[%s] put_message() called with an invalid parameter of type %s",
                         self.interface, type(message))

    def wait_for_single_message(self, timeout: float = None ,
                                keep: Callable[..., bool] = None):
        """Configures the device message queue filter to automatically move messages
        that matches the filter into the queue, and then waits for the first message
        that matches this filter and returns it.
        """
        unexpected_messages = []
        if keep is not None:
            self.set_queue_filter(keep)

        # Wait for a matching message to be caught (blocking)
        msg = self.__out_messages.get(block=True, timeout=timeout)

        # If message filter is set and message does not match, wait until an
        # expected message matches
        if keep is not None:
            # Wait for a matching message
            while not keep(msg):
                unexpected_messages.append(msg)
                msg = self.__out_messages.get(block=True, timeout=timeout)

            # Re-enqueue non-matching messages
            for m in unexpected_messages:
                self.__out_messages.put(m)
        return msg


    def wait_for_message(self, timeout: float = None, keep: Callable[..., bool] = None,
                         command: bool = False):
        """
        Configures the device message queue filter to automatically move messages
        that matches the filter into the queue, and then waits for the first message
        that matches this filter and process it.

        This method is blocking until a matching message is received.

        :param int timeout: Timeout
        :param filter: Message queue filtering function (optional)
        """

        # Raise a WhadDeviceDisconnected exception when the interface is still
        # considered opened but no messages to read in its output message queue.
        #
        # (This specific condition is met when a Unix client socket has closed
        # following a connection to a server, after the server sent a set of
        # messages that are still to process by the client).
        if not self.opened and self.__out_messages.empty():
            logger.debug("[%s] wait_for_message() cannot succeed because device is closed.")
            raise WhadDeviceDisconnected()

        if keep is not None:
            self.set_queue_filter(keep)

        start_time = time()

        while True:
            try:
                # Wait for a matching message to be caught (blocking)
                msg = self.__out_messages.get(block=True, timeout=timeout)

                # If we receive a disconnection event, raise an exception
                if isinstance(msg, Disconnected):
                    raise WhadDeviceDisconnected()

                # If message does not match, re-enqueue for late processing
                if not self.__msg_filter(msg):
                    self.put_message(msg)
                else:
                    # If it does match, return it
                    return msg
            except Empty as err:
                # Queue is empty, wait for a message to show up.
                if timeout is not None and (time() - start_time > timeout):
                    if command:
                        raise WhadDeviceTimeout("WHAD device did not answer to a command") from err

                    logger.debug("exiting wait_for_message (timeout: %s)...", timeout)
                    return None

    def send_message(self, message: HubMessage, keep: Callable[..., bool] = None):
        """
        Serializes a message and sends it to the interface, without waiting
        for an answer. Optionally, you can update the message queue filter
        if you need to wait for specific messages after the message is sent.

        :param Message message: Message to send
        :param keep: Message queue filter function
        """
        if not self.opened and self.__out_messages.empty():
            logger.debug("[%s] Cannot send message: device closed.", self.interface)
            raise WhadDeviceDisconnected()

        # Set message queue filter
        if keep is not None:
            self.set_queue_filter(keep)

        # Enqueue message to transmit to the interface
        self.__in_messages.put(message)


    def send_command(self, command: HubMessage, keep: Callable[..., bool] = None):
        """
        Sends a command and awaits a specific response from the device.
        WHAD commands usualy expect a CmdResult message, if `keep` is not
        provided then this method will by default wait for a CmdResult.

        :param Message command: Command message to send to the device
        :param keep: Message queue filter function (optional)
        :returns: Response message from the device
        :rtype: Message
        """
        # If a queue filter is not provided, expect a default CmdResult
        try:
            if keep is None:
                self.send_message(command, message_filter(CommandResult))
            else:
                self.send_message(command, keep)
        except WhadDeviceError as error:
            # Device error has been triggered, it looks like our device is in
            # an unspecified state, notify user.
            logger.debug("WHAD device in error while sending message: %s", error)
            raise error

        try:
            # Retrieve the first message that matches our filter
            result = self.wait_for_message(self.__timeout, command=True)
        except WhadDeviceTimeout as timedout:
            # Forward exception
            raise timedout

        # Log message
        logger.debug("Command result: %s", result)

        # Return command result
        return result


    ##
    # Interface management
    ##

    ######################################
    # Generic discovery
    ######################################

    def on_discovery_msg(self, message):
        """
        Method called when a discovery message is received. If a connector has
        been associated with the device, forward this message to this connector.
        """

        # Forward everything to the connector, if any
        if self.__connector is not None:
            self.__connector.on_discovery_msg(message)

    def has_domain(self, domain) -> bool:
        """Checks if device supports a specific domain.

        :param Domain domain: Domain
        :returns: True if domain is supported, False otherwise.
        :rtype: bool
        """
        if self.__info is not None:
            return self.__info.has_domain(domain)

        # No info available on device, domain is not supported by default
        return False


    def get_domains(self):
        """Get device' supported domains.

        :returns: list of supported domains
        :rtype: list
        """
        if self.__info is not None:
            return self.__info.domains

        # No domain discovered yet
        return []


    def get_domain_capability(self, domain) -> int:
        """Get a device domain capabilities.

        :param Domain domain: Target domain
        :returns: Domain capabilities
        :rtype: DeviceDomainInfoResp
        """
        if self.__info is not None:
            cap = self.__info.get_domain_capabilities(domain)
            if cap is not None:
                return cap

        # No capability if not discovered
        return 0

    def get_domain_commands(self, domain):
        """Get a device supported domain commands.

        :param Domain domain: Target domain
        :returns: Bitmask of supported commands
        :rtype: int
        """
        if self.__info is not None:
            return self.__info.get_domain_commands(domain)

        # No supported commands by default
        return 0


    def send_discover_info_query(self, proto_version=0x0100):
        """
        Sends a DeviceInfoQuery message and awaits for a DeviceInfoResp
        answer.
        """
        logger.info("preparing a DeviceInfoQuery message")
        msg = self.__hub.discovery.create_info_query(proto_version)
        return self.send_command(
            msg,
            message_filter(InfoQueryResp)
        )


    def send_discover_domain_query(self, domain):
        """
        Sends a DeviceDomainQuery message and awaits for a DeviceDomainResp
        answer.
        """
        logger.info("preparing a DeviceDomainInfoQuery message")
        msg = self.__hub.discovery.create_domain_query(domain)
        return self.send_command(
            msg,
            message_filter(DomainInfoQueryResp)
        )

    def discover(self):
        """
        Performs device discovery (synchronously).

        Discovery process asks the device to provide its description, including
        its supported domains and associated capabilities. For each domain we
        then query the device and get the list of supported commands.
        """
        if not self.__discovered:
            # We send a DeviceInfoQuery message to the device and expect a
            # DeviceInfoResponse in return.
            resp = self.send_discover_info_query()

            # If we have an answer, process it.
            if resp is not None:

                # Ensure response is the one we expect
                assert isinstance(resp, InfoQueryResp)

                # Save device information
                self.__info = DeviceInfo(
                    resp
                )

                # Parse DeviceInfoResponse
                #device_info = self.hub.parse(resp)

                # Update our ProtocolHub version to the device version
                self.__hub = ProtocolHub(resp.proto_min_ver)

                # Query device domains
                logger.info("query supported commands per domain")
                for domain in self.__info.domains:
                    resp = self.send_discover_domain_query(domain)
                    self.__info.add_supported_commands(
                        resp.domain,
                        resp.supported_commands
                    )

                # Mark device as discovered
                logger.info("device discovery done")
                self.__discovered = True

                # Switch to max transport speed
                logger.info("set transport speed to %d", self.info.max_speed)
                self.change_transport_speed(
                    self.info.max_speed
                )
            else:
                logger.error("device is not ready !")
                raise WhadDeviceNotReady()

    def reset(self):
        """Reset device
        """
        logger.info("preparing a DeviceResetQuery message")
        msg = self.__hub.discovery.create_reset_query()
        return self.send_command(
            msg,
            message_filter(DeviceReady)
        )

class VirtualDevice(Device):
    """
    Virtual interface implementation.

    This variant of the base Interface class provides a way to emulate an interface
    compatible with WHAD. This emulated compatible interface is used as an adaptation
    layer between WHAD's core and third-party hardware that does not run a WHAD-enabled
    firmware.
    """
    def __init__(self, index: Optional[int] = None):
        self._dev_type = None
        self._dev_id = None
        self._fw_author = None
        self._fw_url = None
        self._fw_version = (0, 0, 0)
        self._dev_capabilities = {}
        self.__lock = Lock()
        super().__init__(index)

    def send_message(self, message, keep=None):
        """Send message to host.
        """
        logger.debug("[virtual_iface][%s] send_message(%s)", self.interface, message)
        with self.__lock:
            super().set_queue_filter(keep)
            self._on_whad_message(message)

    def _on_whad_message(self, message):
        """TODO: associate callbacks with classes ?
        """
        logger.debug("on_whad_message: %s", message)
        category = message.message_type
        message_type = message.message_name

        callback_name = f"_on_whad_{category}_{message_type}"
        if hasattr(self, callback_name) and callable(getattr(self, callback_name)):
            getattr(self, callback_name)(message)
        else:
            logger.info("unhandled message: %s", message)
            self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_discovery_info_query(self, _):
        major, minor, revision = self._fw_version
        msg = self.hub.discovery.create_info_resp(
            DeviceType.VirtualDevice,
            self._dev_id,
            0x0100,
            0,
            self._fw_author,
            self._fw_url,
            major, minor, revision,
            [dom | (cap[0] & 0xFFFFFF) for dom, cap in self._dev_capabilities.items()]
        )
        self._send_whad_message(msg)

    def _on_whad_discovery_domain_query(self, message):
        # Compute supported commands for domain
        commands = 0
        supported_commands = self._dev_capabilities[message.domain][1]
        for command in supported_commands:
            commands |= (1 << command)

        # Create a DomainResp message and send it
        msg = self.hub.discovery.create_domain_resp(
            message.domain,
            commands
        )
        self._send_whad_message(msg)


    def _send_whad_message(self, message):
        self.put_message(message)

    def _send_whad_command_result(self, code):
        msg = self.hub.generic.create_command_result(code)
        self._send_whad_message(msg)


class WhadDevice(Device):
    """
    This class is an alias for :py:class:`whad.device.device.Device`,
    and will be deprecated in a near future. This class has been introduced
    in a previous version of WHAD and has been renamed for clarity purpose.
    """

    @classmethod
    def create_inst(cls, interface_string):
        """Create an instance of interface from its name, using Device."""
        return Device.create_inst(interface_string)

    @classmethod
    def create(cls, interface_string):
        """Create an instance of interface from its name, using Device."""
        return Device.create(interface_string)

    @classmethod
    def check_interface(cls, interface):
        """Check if Device supports the requested interface."""
        return Device.check_interface(interface)

class WhadVirtualDevice(VirtualDevice):
    """
    This class is an alias for :py:class:`whad.device.device.VirtualDevice`,
    and will be deprecated in a near future. This class has been introduced
    in a previous version of WHAD and has been renamed for clarity purpose.
    """

    @classmethod
    def create_inst(cls, interface_string):
        """Create an instance of interface from its name, using VirtualDevice."""
        return VirtualDevice.create_inst(interface_string)

    @classmethod
    def create(cls, interface_string):
        """Create an instance of interface from its name, using VirtualDevice."""
        return VirtualDevice.create(interface_string)

    @classmethod
    def check_interface(cls, interface):
        """Check if VirtualDevice supports the requested interface."""
        return VirtualDevice.check_interface(interface)




"""Logitech Unifying conenctors
"""
from .base import Unifying, ESBAddress
from .sniffer import Sniffer
from .keylogger import Keylogger
from .mouselogger import Mouselogger
from .mouse import Mouse
from .keyboard import Keyboard
from .dongle import Dongle
from .injector import Injector

__all__ = [
    "ESBAddress",
    "Unifying",
    "Sniffer",
    "Keylogger",
    "Mouselogger",
    "Mouse",
    "Keyboard",
    "Dongle",
    "Injector",
]



"""BLE utility device caching module.

This module implements a single class, `BleDevicesCache`, that keeps track
of discovered devices and their information.
"""
from whad.ble.scanning import AdvertisingDevice
from whad.ble.profile.device import PeripheralDevice

class BleDevicesCache:
    """Bluetooth Low Energy devices cache.
    """

    def __init__(self):
        """Initialize device cache.
        """
        self.__devices = {}

    def __len__(self):
        """Retrieve cache length
        """
        return len(self.__devices.keys())

    def clear(self):
        """Remove device entry from cache.
        """
        self.__devices = {}

    def iterate(self):
        """Iterate over cached devices.
        """
        for _, device in self.__devices.items():
            yield device

    def add(self, device: AdvertisingDevice):
        """Add a discovered device to our cache.

        @param device AdvertisingDevice Device to add
        """
        self.__devices[device.address.lower()] = {
            'info':device,
            'profile': None,
            'discovered': False
        }

    def __getitem__(self, address: str):
        """Return an existing device from cache.

        @param address str Device BD address (i.e. '00:11:22:33:44:55')
        """
        if address in self.__devices:
            return self.__devices[address.lower()]

        # If not found, look into devices names
        for _, dev in self.__devices.items():
            if dev['info'].name is not None and dev['info'].name == address:
                return dev
        raise IndexError

    def add_profile(self, address: str, profile: PeripheralDevice):
        """Add profile to a cached device.

        @param address str: Device BD address
        @param profile PeripheralDevice: Device object (obtained when connected)
        """
        address = address.lower()
        if address in self.__devices:
            self.__devices[address.lower()]['profile'] = profile
        else:
            raise IndexError

    def get_profile(self, address: str):
        """Get cached services associated with a specific device.
        """
        address = address.lower()
        if address in self.__devices:
            # Device found, check if we have a cached profile
            return self.__devices[address]['profile']

        # Not found
        raise IndexError

    def mark_as_discovered(self, address: str):
        """Mark device as discovered.
        """
        address = address.lower()
        if address in self.__devices:
            # Device found, check if we have a cached profile
            self.__devices[address]['discovered'] = True
        else:
            raise IndexError

    def is_discovered(self, address) -> bool:
        """Determine if a device has already been discovered.
        """
        address = address.lower()
        if address in self.__devices:
            # Device found, check if we have a cached profile
            return self.__devices[address]['discovered']

        # Device is not in our database
        raise IndexError



"""This module provides different classes that represent a BLE device and
allows to interact with it:

* :class:`whad.ble.profile.Profile` is a base class used to register all
  the ATT attributes, including services, characteristics, characteristic values
  and descriptors. It is able to inspect any derived class and build the
  corresponding profile based on properties declared with
  :class:`whad.ble.profile.service.PrimaryService` and
  :class:`whad.ble.profile.characteristic.Characteristic`.

* :class:`whad.ble.profile.read` is a decorator class used to mark a method
  as a callback associated to a GATT read operation for a specific characteristic.
* :class:`whad.ble.profile.write` is a decorator class used to mark a method as a
  callback associated to a GATT write operation to be performed on a specific characteristic.
* :class:`whad.ble.profile.written` is a decorator class used to mark a methiod as
  a callback associated to a performed GATT write operation for a specific characteristic.
* :class:`whad.ble.profile.subscribed` is a decorator class used to mark a method as
  a callback associated to a subscription for notification or indication for a specific
  characteristic
* :class:`whad.ble.profile.unsubscribed` is a decorator class used to mark a method as
  a callback associated to an unsubscription for notification or indication for a specific
  characteristic
"""
import json
import logging
from typing import List, Iterator, Optional, Callable, Any, Union

from whad.ble.profile.attribute import Attribute, UUID
from whad.ble.profile.characteristic import (
    Characteristic, CharacteristicValue, Properties, ClientCharacteristicConfig, Descriptor, CharacteristicDescriptor,
    ReportReference, UserDescription, CharacteristicUserDescriptionDescriptor, ReportReferenceDescriptor
)

from whad.ble.profile.service import PrimaryService, SecondaryService, IncludeService, Service
from whad.ble.exceptions import InvalidHandleValueException
from whad.ble.stack.att.constants import SecurityAccess

logger = logging.getLogger(__name__)

###################################
# Decorators for GenericProfile
###################################

class CharacteristicHook:
    """Characteristic hook decorator.

    This class defines the base hook decorator class and must be inherited.
    """

    def __init__(self, *args):
        if len(args) == 1:
            characteristic = args[0]
            if isinstance(characteristic, Characteristic):
                self.__characteristic = f"{characteristic.service.uuid}:{characteristic.uuid}"
            else:
                raise TypeError
        elif len(args) == 2:
            service = args[0]
            charac = args[1]
            if isinstance(service, str) and isinstance(charac, str):
                self.__characteristic = f"{UUID(service)}:{UUID(charac)}"
            elif isinstance(service, UUID) and isinstance(charac, UUID):
                self.__characteristic = f"{service}:{charac}"

    def __call__(self, method):
        if not hasattr(method, "hooks"):
            method.hooks = []
        if not hasattr(method, "characteristic"):
            method.characteristic = self.__characteristic
        return method

# Decorator name does not start with a capital letter for ease of use
# pylint: disable-next=invalid-name
class read(CharacteristicHook):
    """Read hook decorator

    This decorator is used to declare a callback method for read operations
    on a specific characteristic.
    """

    def __call__(self, method):
        """Add a specific hook to the method
        """
        super().__call__(method)
        if "read" not in method.hooks:
            method.hooks.append('read')
        return method

# Decorator name does not start with a capital letter for ease of use
# pylint: disable-next=invalid-name
class write(CharacteristicHook):
    """Write hook decorator

    This decorator is used to declare a callback method for write operations
    on a specific characteristic. This callback will be called **before** the
    write operation happens.
    """

    def __call__(self, method):
        """Add a specific hook to the method
        """
        super().__call__(method)
        if "write" not in method.hooks:
            method.hooks.append('write')
        return method

# Decorator name does not start with a capital letter for ease of use
# pylint: disable-next=invalid-name
class written(CharacteristicHook):
    """Written hook decorator

    This decorator is used to declare a callback method for write operations
    on a specific characteristic. This callback will be called **after** the
    write operation happens.
    """

    def __call__(self, method):
        """Add a specific hook to the method
        """
        super().__call__(method)
        if "written" not in method.hooks:
            method.hooks.append("written")
        return method

# Decorator name does not start with a capital letter for ease of use
# pylint: disable-next=invalid-name
class subscribed(CharacteristicHook):
    """Subscribe hook decorator

    This decorator is used to declare a callback method for subscribe operations
    on a specific characteristic.
    """

    def __call__(self, method):
        """Add a specific hook to the method
        """
        super().__call__(method)
        if "sub" not in method.hooks:
            method.hooks.append("sub")
        return method

# Decorator name does not start with a capital letter for ease of use
# pylint: disable-next=invalid-name
class unsubscribed(CharacteristicHook):
    """Unsubscribe hook decorator

    This decorator is used to declare a callback method for unsubscribe operations
    on a specific characteristic.
    """

    def __call__(self, method):
        """Add a specific hook to the method
        """
        super().__call__(method)
        if "unsub" not in method.hooks:
            method.hooks.append("unsub")
        return method

def is_method_hook(method):
    """Determine if a method is a characteristic operation hook
    """
    if hasattr(method, "hooks") and hasattr(method, "characteristic"):
        return len(method.hooks) > 0
    return False

class Profile:
    """This class implements a GATT profile, i.e. a set of services and characteristics
    exposed by a Bluetooth Low Energy GATT server.
    """

    def __init__(self, start_handle: int = 1, from_json: Optional[str] = None):
        """Parse the device model, instantiate all the services, characteristics
        and descriptors, compute all handle values and registers everything
        inside this instance for further use.

        :param  start_handle:   Start handle value to use (default: 0)
        :type   start_handle:   int
        :param  from_json:      JSON data describing a GATT profile
        :type   from_json:      str
        """
        self.__attr_db = {}
        self.__services = []
        self.__service_by_characteristic_handle = {}

        self.__start_handle = start_handle
        self.__handle = self.__start_handle

        self.__hooks = {}

        # Populate attribute database and model from JSON export if provided
        if from_json is not None:
            from_json: dict = json.loads(from_json)
            # Parse JSON services, characteristics and descriptors to create
            # the corresponding model and attribute database
            if 'services' in from_json:
                # Loop on services
                for service in from_json['services']:
                    # Collect characteristics
                    if UUID(service['type_uuid']) == UUID(0x2800):
                        service_obj = PrimaryService(
                            uuid=UUID(service['uuid']),
                            handle=service['start_handle'],
                            end_handle=service['end_handle'],
                        )
                    elif UUID(service['type_uuid']) == UUID(0x2801):
                        service_obj = SecondaryService(
                            uuid=UUID(service['uuid']),
                            handle=service['start_handle'],
                            end_handle=service['end_handle']
                        )
                    else:
                        # This is not a known service type UUID, continue with
                        # next service
                        continue

                    if 'characteristics' in service:
                        for charac in service['characteristics']:
                            # Load characteristic data (if provided)
                            charac_data = b''
                            if 'data' in charac['value']:
                                charac_data = bytes.fromhex(charac['value']['data'])

                            # Create characteristic model
                            charac_obj = Characteristic(
                                uuid=UUID(charac['value']['uuid']),
                                handle=charac['handle'],
                                value=charac_data,
                                properties=charac['properties'],
                                security=SecurityAccess.int_to_accesses(charac['security'])
                            )

                            # Loop on descriptors for the current characteristic
                            for desc in charac['descriptors']:
                                # Try to convert this descriptor to an instance
                                # of one of our supported descriptors
                                desc_obj = Descriptor.from_uuid(
                                    handle=desc['handle'],
                                    uuid=UUID(desc['uuid']),
                                    value=bytes.fromhex(desc['value']) if 'value' in desc else b'',
                                    characteristic=charac_obj,
                                )

                                # Add descriptor
                                if desc_obj is not None:
                                    charac_obj.add_descriptor(desc_obj)
                                    self.register_attribute(desc_obj)

                            # Register characteristic and its value
                            self.register_attribute(charac_obj)
                            self.register_attribute(charac_obj.value_attr)

                            # Add characteristic to its related service
                            service_obj.add_characteristic(charac_obj)

                    self.register_attribute(service_obj)
                    self.add_service(service_obj)
        else:
            # Introspect this class definition and build model
            services = []
            props = dir(self)
            for prop in props:
                if not prop.startswith('_'):
                    if isinstance(getattr(self, prop), Service):
                        service = getattr(self, prop)
                        services.append((prop, service))

            # Instantiate each service, and for each of them the corresponding
            # characteristics
            for name, service in services:
                if isinstance(service, PrimaryService):
                    logger.info("creating primary service %s", service.uuid)
                    # Create service
                    service_obj = service.build()

                elif isinstance(service, SecondaryService):
                    logger.info("creating secondary service %s", service.uuid)
                    # Create service
                    service_obj = service.build()
                else:
                    continue

                # Overwrite the corresponding instance property with our new
                # service instance.
                self.add_service(service_obj)
                setattr(self, name, service_obj)

            # We then need to update included service start and end handles
            for inc_service in self.included_services():
                # Retrieve the included service UUID
                service_uuid = inc_service.service_uuid

                # Retrieve the corresponding object by UUID
                service_obj = self.get_service_by_uuid(service_uuid)

                # If found, update start and end handles
                if service_obj is not None and service_obj.handle is not None:
                    inc_service.service_start_handle = service_obj.handle
                    inc_service.service_end_handle = service_obj.end_handle


        # Register any hook function declared in profile class
        props = dir(self)
        for prop in props:
            prop_obj = getattr(self, prop)
            # Is this property a callable hook ?
            if callable(prop_obj) and is_method_hook(prop_obj):
                # Associate hook method with each operation
                if prop_obj.characteristic not in self.__hooks:
                    self.__hooks[prop_obj.characteristic] = {}
                for operation in prop_obj.hooks:
                    self.__hooks[prop_obj.characteristic][operation] = prop_obj


    @property
    def db(self):
        return self.__attr_db

    def __alloc_handle(self, number: int = 1):
        """Allocate one or more handle values.

        :param  number: Number of handle values to allocate
        :type   number: int

        :return: Current handle value
        :rtype: int
        """
        current_handle = self.__handle
        self.__handle += number
        return current_handle

    def __repr__(self):
        output = ''
        for service in self.services():
            output += (
                f"Service {service.uuid} (handles from {service.handle:d} to "
                f"{service.end_handle:d}):\n"
            )

            for inc_service in service.included_services():
                output += (
                    f"  Included service {inc_service.service_uuid} "
                    f"(handle:{inc_service.handle:d}, "
                    f"start_handle:{inc_service.service_start_handle:d}, "
                    f"end_handle:{inc_service.service_end_handle:d})\n"
                )
            for charac in service.characteristics():
                properties = charac.properties
                charac_rights = ''
                if properties & Properties.READ != 0:
                    charac_rights += 'R'
                if properties & Properties.WRITE != 0:
                    charac_rights += 'W'
                if properties & Properties.INDICATE != 0:
                    charac_rights += 'I'
                if properties & Properties.NOTIFY != 0:
                    charac_rights += 'N'

                output += (
                    f"  Characteristic {charac.uuid} (handle:{charac.handle:d}, "
                    f"value handle: {charac.value_handle:d}, "
                    f"props: {charac_rights}, {charac.alias})\n"
                )
                for desc in charac.descriptors():
                    output += f"    Descriptor {desc.type_uuid} (handle: {desc.handle:d})\n"
        return output

    def register_attribute(self, attribute: Attribute):
        """Register a GATT attribute

        :param  attribute:  Attribute to register
        :type   attribute:  Attribute
        """
        if isinstance(attribute, Attribute):
            self.__attr_db[attribute.handle] = attribute


    def add_service(self, service: Service, handles_only: bool = False):
        """Add a service to the current device

        :param  service:        Service to add to the device
        :type   service:        Service
        :param  handles_only:   Add only service handles if set to ``True``
        :type   handles_only:   bool
        """
        logger.debug("add service %s", service.uuid)
        if service.handle == 0:
            # Service has not been fully configured, update its handle
            # and the handles of its characteristics and descriptors.
            service.handle = self.__alloc_handle()

        # Append service to the list of our services
        if not handles_only and service not in self.__services:
            self.__services.append(service)

        # Register service as an attribute
        self.register_attribute(service)

        # Register all its characteristics
        for charac in service.characteristics():
            # Register Characteristic and its CharacteristicValue
            self.register_attribute(charac)
            self.register_attribute(charac.value_attr)

            # Register characteristic's descriptors
            for desc in charac.descriptors():
                self.register_attribute(desc)

            # Add characteristic in our lookup table
            self.__service_by_characteristic_handle[charac.handle] = service

        # Update our last handle based on service's end handle
        self.__handle = service.end_handle + 1

    def remove_service(self, service: Service, handles_only: bool = False):
        """Remove service

        :param  service:        Service object or UUID
        :type   service:        Service
        :param  handles_only:   Remove only handles if set to ``True``
        :type   handles_only:   bool
        """
        if isinstance(service, (PrimaryService, SecondaryService)):
            service_obj = self.get_service_by_uuid(service.uuid)
        elif isinstance(service, UUID):
            service_obj = self.get_service_by_uuid(service)
        else:
            service_obj = None

        # Process service object
        if service_obj is not None:
            # Remove service and all its characteristics from the attribute DB
            for charac in service_obj.characteristics():
                # Remove characteristic handle
                if charac.handle in self.__attr_db:
                    del self.__attr_db[charac.handle]

                # Remove characteristic value handle
                if charac.value_handle in self.__attr_db:
                    del self.__attr_db[charac.value_handle]

                # Remove all the attached descriptors
                for desc in charac.descriptors():
                    if desc.handle in self.__attr_db:
                        del self.__attr_db[desc.handle]

            # Remove service object from attribute db
            del self.__attr_db[service_obj.handle]

            # Remove service from our list of services (if required)
            if not handles_only:
                self.__services.remove(service)
        else:
            # Not found, raise IndexError
            raise IndexError()


    def update_service(self, service: Service) -> bool:
        """Update service in profile.

        Keep service in place in the service list,
        but update all the services declared after this one.

        :param  service:    Service object to update.
        :type   service:    Service
        :return: ``True`` if service has been updated, ``False`` otherwise.
        :rtype: bool
        """
        try:
            service_index = self.__services.index(service)

            # Remove all handles used by this service
            self.remove_service(service, handles_only=True)

            # Register all the handles back into our attribute DB
            self.add_service(service, handles_only=True)

            # Update all other services
            handle = service.end_handle
            for remaining_service in self.__services[service_index+1:]:
                remaining_service.handle = handle + 1
                self.update_service(remaining_service)
                handle = remaining_service.end_handle
            self.__handle = handle
            return True
        except IndexError:
            return False

    def find_object_by_handle(self, handle: int) -> Attribute:
        """Find an object by its handle value

        :param  handle: Object handle
        :type   handle: int
        :return: Object if handle is valid, or raise an IndexError exception otherwise
        :rtype: Attribute
        :raises: IndexError
        """
        if handle in self.__attr_db:
            return self.__attr_db[handle]

        # Error.
        raise IndexError

    def find_objects_by_range(self, start: int, end: int) -> List[Attribute]:
        """Find attributes with handles belonging in the [start, end+1] interval.

        :param  start:  Start handle value
        :type   start:  int
        :param  end:    End handle value
        :type   end:    int
        :return:        List of objects with handles between start and end values
        :rtype: list
        :raises: IndexError
        """
        handles = []
        for handle in self.__attr_db:
            if start <= handle <= end:
                handles.append(handle)
        handles.sort()
        return [self.find_object_by_handle(handle) for handle in handles]


    def find_characteristic_by_value_handle(self, value_handle: int) -> Optional[Characteristic]:
        """Find characteristic object by its value handle.

        :param  value_handle:   Characteristic value handle
        :type   value_handle:   int
        :return: Corresponding characteristic object or ``None`` if not found.
        :rtype: Characteristic
        """
        try:
            char_value = self.find_object_by_handle(value_handle)
            if isinstance(char_value, CharacteristicValue) and hasattr(char_value, 'characteristic'):
                return char_value.characteristic

            # Not found.
            return None
        except InvalidHandleValueException:
            return None


    def find_characteristic_end_handle(self, handle: int) -> Optional[int]:
        """Find characteristic end handle based on its handle.

        :param  handle: Characteristic handle
        :type   handle: int
        :rtype: int
        :return: Characteristic value handle
        :raises: InvalidHandleValueException
        """
        try:
            # Find service owning the characteristic
            service = self.find_service_by_characteristic_handle(handle)

            # Build a list of characteristic handles
            service_char_handles=[]
            for characteristic in service.characteristics():
                service_char_handles.append(characteristic.handle)

            # Sort handles
            service_char_handles.sort()
            idx = service_char_handles.index(handle)
            if idx == len(service_char_handles) - 1:
                return service.end_handle

            return service_char_handles[idx+1] - 1

        except InvalidHandleValueException:
            return None


    def find_service_by_characteristic_handle(self, handle: int) -> Service:
        """Find a service object given a characteristic handle that belongs
        to this service.

        :param  handle: Characteristic handle belonging to the searched service
        :type   handle: int
        :rtype: Service
        :return: Service object containing the specified characteristic

        :raises: InvalidHandleValueException
        """
        try:
            if handle in self.__service_by_characteristic_handle:
                return self.__service_by_characteristic_handle[handle]

            # Invalid handle
            raise InvalidHandleValueException
        except IndexError as err:
            raise InvalidHandleValueException from err


    def services(self) -> Iterator[Service]:
        """Enumerate service objects.

        This method is a generator and will yield service objects registered
        into the profile.
        """
        for _, obj in self.__attr_db.items():
            if isinstance(obj, Service):
                yield obj

    def included_services(self) -> Iterator[IncludeService]:
        """Enumerate included services.
        """
        for _, obj in self.__attr_db.items():
            if isinstance(obj, IncludeService):
                yield obj

    def service(self, uuid: Union[str, UUID]) -> Optional[Service]:
        """Retrieve a Service object given its UUID.

        :param uuid:    Service UUID
        :type  uuid:    UUID, str
        :return:        Corresponding Service object if found, ``None`` otherwise.
        :rtype:         Service
        :raise:         InvalidUUIDException
        """
        # If a string is provided as UUID, convert it to the corresponding
        # UUID object. This could raise an InvalidUUIDException.
        if isinstance(uuid, str):
            uuid = UUID(uuid)

        # Search for a service matching the given UUID
        for _, obj in self.__attr_db.items():
            if isinstance(obj, (PrimaryService, SecondaryService)):
                if obj.uuid == uuid:
                    return obj

        # Not found
        return None

    def get_service_by_uuid(self, uuid: Union[str, UUID]) -> Optional[Service]:
        """Retrieve a Service object given its UUID.

        :param uuid:    Service UUID
        :type  uuid:    UUID
        :type  uuid:    str
        :return:        Corresponding Service object if found, `None` otherwise.
        :rtype:         Service, optional
        :raises InvalidUUIDException: Specified UUID is invalid

        .. deprecated:: 1.3.0
            Use the :py:meth:`~whad.ble.profile.service` method to find a service
            based on its UUID (simpler syntax).
        """
        return self.service(uuid)

    def char(self, uuid: Union[str, UUID]) -> Optional[Characteristic]:
        """Get characteristic by its UUID.

        :param      uuid:   Characteristic UUID to look for
        :type       uuid:   :class:`whad.ble.profile.attribute.UUID`
        :type       uuid:   str
        :return:            Characteristic if found, ``None`` otherwise
        :rtype:             :class:`whad.ble.profile.characteristic.Characteristic`, optional
        """
        # If a string is provided as UUID, convert it to the corresponding
        # UUID object. This could raise an InvalidUUIDException.
        if isinstance(uuid, str):
            uuid = UUID(uuid)

        # Search for a characteristic with the given UUID
        for _, obj in self.__attr_db.items():
            if isinstance(obj, Characteristic):
                if obj.uuid == uuid:
                    return obj

        # Not found
        return None

    def get_characteristic_by_uuid(self, uuid: Union[str, UUID]):
        """Get characteristic by its UUID.

        :param      uuid:   Characteristic UUID to look for
        :type       uuid:   :class:`whad.ble.profile.attribute.UUID`
        :type       uuid:   str
        :return:            Characteristic if found, ``None`` otherwise
        :rtype:             :class:`whad.ble.profile.characteristic.Characteristic`
        :raises InvalidUUIDException: Specified UUID is invalid
        """
        return self.char(uuid)

    def attr_by_type_uuid(self, uuid, start: int = 1, end: int = 0xFFFF) -> Iterator[Attribute]:
        """Enumerate attributes that have a specific type UUID.

        :param  uuid:   Type UUID
        :type   uuid:   UUID
        :param  start:  Start handle
        :type   start:  int
        :param  end:    End handle
        :type   end:    int
        """
        for _, obj in self.__attr_db.items():
            if obj.type_uuid == uuid and start <= obj.handle <= end:
                yield obj

    def export_json(self) -> str:
        """Export profile as JSON data, including services, characteristics and descriptors
        definition.

        :return:    JSON data corresponding to this profile
        :rtype:     str
        """
        profile_dict = {}
        profile_dict['services'] = []
        for service in self.services():
            service_dict = {
                'uuid': str(service.uuid),
                'type_uuid': str(service.type_uuid),
                'start_handle': service.handle,
                'end_handle': service.end_handle
            }
            service_dict['characteristics'] = []
            for charac in service.characteristics():
                charac_dict = {
                    'handle': charac.handle,
                    'uuid': str(charac.type_uuid),
                    'properties': charac.properties,
                    'security': SecurityAccess.accesses_to_int(charac.security),
                    'value': {
                        'handle': charac.value_handle,
                        'uuid': str(charac.uuid),
                        'data': Attribute.value.fget(charac.value_attr).hex()
                    }
                }
                charac_dict['descriptors'] = []
                for desc in charac.descriptors():
                    desc_dict = {
                        'handle': desc.handle,
                        'uuid': str(desc.type_uuid),
                        'value': Attribute.value.fget(desc).hex()
                    }
                    charac_dict['descriptors'].append(desc_dict)
                service_dict['characteristics'].append(charac_dict)
            profile_dict['services'].append(service_dict)
        return json.dumps(profile_dict)


    def find_hook(self, service: Service, characteristic: Characteristic,
                  operation: str) -> Optional[Callable[..., Any]]:
        """Find a registered hook for a specific service, characteristic and operation.

        :param  service:        Service object
        :type   service:        Service
        :param  characteristic: Characteristic object
        :type   characteristic: Characteristic
        :param  operation:      GATT operation
        :type   operation:      str

        :return: Hook callback
        :rtype: callable
        """
        hook_key = str(service.uuid) + ':' + str(characteristic.uuid)
        if hook_key in self.__hooks:
            if operation in self.__hooks[hook_key]:
                return self.__hooks[hook_key][operation]
        return None


    ################################################
    # Connection/disconnection hooks
    ################################################

    def on_connect(self, conn_handle: int):
        """Connection hook.

        This hook is only used to notify the connection of a device.

        :param  conn_handle:    Connection handle
        :type   conn_handle:    int
        """
        logger.debug("[profile] Connection established with handle %d",
                     conn_handle)

    def on_disconnect(self, conn_handle: int):
        """Disconnection hook.

        This hook is only used to notify the disconnection of a device.

        :param  conn_handle:    Connection handle
        :type   conn_handle:    int
        """
        logger.debug("[profile] Connection terminated for handle %d",
                     conn_handle)


    ################################################
    # Characteristic Read/Write/Subscribe hooks
    ################################################

    def on_characteristic_read(self, service: Service, characteristic: Characteristic,
                               offset: int = 0, length: int = 0):
        """Characteristic read hook.

        This hook is called whenever a characteristic is about to be read by a GATT client.
        If this method returns a byte array, this byte array will be sent back to the
        GATT client. If this method returns None, then the read operation will return an
        error (not allowed to read characteristic value).


        :param  service:        Service owning the characteristic
        :type   service:        Service
        :param  characteristic: Characteristic object
        :type   characteristic: Characteristic
        :param  offset:         Read offset (default: 0)
        :type   offset:         int
        :param  length:         Max read length
        :type   length:         int

        :return:    Value to return to the GATT client
        :rtype:     bytes
        """
        # Check if we have a hook to call
        hook = self.find_hook(service, characteristic, 'read')
        if hook is not None:
            return hook(offset, length)

        # If no hook registered, then return the characteristic value
        return characteristic.value[offset:offset + length]

    def on_characteristic_write(self, service: Service, characteristic: Characteristic,
                                offset: int = 0, value: bytes = b'',
                                without_response: bool = False):
        """Characteristic write hook

        This hook is called whenever a charactertistic is about to be written by a GATT
        client.

        :param  service:            Service owning the characteristic
        :type   service:            Service
        :param  characteristic:     Characteristic object
        :type   characteristic:     Characteristic
        :param  offset:             Read offset (default: 0)
        :type   offset:             int
        :param  value:              Value about to be written into the characteristic
        :type   value:              bytes
        :param  without_response:   Set to ``True`` if no response is required
        :type   without_response:   bool
        """
        hook = self.find_hook(service, characteristic, 'write')
        if hook is not None:
            return hook(
                offset,
                value,
                without_response=without_response
            )

        # No action
        return None

    def on_characteristic_written(self, service: Service, characteristic: Characteristic,
                                  offset: int = 0, value: bytes = b'',
                                  without_response: bool = False):
        """Characteristic written hook

        This hook is called whenever a charactertistic has been written by a GATT
        client.

        :param  service:            Service owning the characteristic
        :type   service:            Service
        :param  characteristic:     Characteristic object
        :type   characteristic:     Characteristic
        :param  offset:             Read offset (default: 0)
        :type   offset:             int
        :param  value:              Value about to be written into the characteristic
        :type   value:              bytes
        :param  without_response:   Set to ``True`` if no response is required
        :type   without_response:   bool
        """
        # Check if we have a hook to call
        hook = self.find_hook(service, characteristic, 'written')
        if hook is not None:
            # Call our hook
            return hook(
                offset,
                value,
                without_response=without_response
            )

        # No action
        return None

    def on_characteristic_subscribed(self, service: Service, characteristic: Characteristic,
                                     notification: bool = False, indication: bool = False):
        """Characteristic subscribed hook

        This hook is called whenever a characteristic has been subscribed to.

        :param  service:            Service owning the characteristic
        :type   service:            Service
        :param  characteristic:     Characteristic object
        :type   characteristic:     Characteristic
        :param  notification:       Set to ``True`` if subscribed to notification
        :type   notification:       bool
        :param  indication:         Set to ``True`` if subscribed to notification
        :type   indication:         bool
        """
        # Check if we have a hook to call
        hook = self.find_hook(service, characteristic, 'sub')
        if hook is not None:
            # Call our hook
            return hook(
                notification=notification,
                indication=indication
            )

        # No action
        return None

    def on_characteristic_unsubscribed(self, service: Service, characteristic: Characteristic):
        """Characteristic unsubscribed hook

        This hook is called whenever a characteristic has been unsubscribed.

        :param  service:            Service owning the characteristic
        :type   service:            Service
        :param  characteristic:     Characteristic object
        :type   characteristic:     Characteristic
        """
        # Check if we have a hook to call
        hook = self.find_hook(service, characteristic, 'unsub')
        if hook is not None:
            # Call our hook
            return hook()

        # No action
        return None

    def on_notification(self, service: Service, characteristic: Characteristic, value: bytes):
        """Characteristic notification hook.

        This hook is called when a notification is sent to a characteristic.

        :param  service:            Service owning the characteristic
        :type   service:            Service
        :param  characteristic:     Characteristic object
        :type   characteristic:     Characteristic
        :param  value:              Characteristic value
        :type   value:              bytes
        """
        logger.debug("[profile] GATT notification sent for characteristic %s of service %s with value %s",
                     characteristic.uuid, service.uuid,value)

    def on_indication(self, service: Service, characteristic: Characteristic, value: bytes):
        """Characteristic indication hook.

        This hook is called when a indication is sent to a characteristic.

        :param  service:            Service owning the characteristic
        :type   service:            Service
        :param  characteristic:     Characteristic object
        :type   characteristic:     Characteristic
        :param  value:              Characteristic value
        :type   value:              bytes
        """
        logger.debug("[profile] GATT indication sent for characteristic %s of service %s with value %s",
                     characteristic.uuid, service.uuid,value)

    def on_mtu_changed(self, mtu: int):
        """MTU change callback

        :param  mtu: New MTU value
        :type   mtu: int
        """
        logger.debug("[profile] GATT MTU updated to %d", mtu)

class GenericProfile(Profile):
    """Old name of the `Profile` class, kept for backward compatibility.

    .. versionchanged:: 1.3.0
        :class:`~whad.ble.profile.GenericProfile` has been renamed to :class:`~whad.ble.profile.Profile` to simplify
        code and due to a change in the way standard services are now declared within a GATT profile class.

        In previous versions, including a *Battery Service* into a custom profile required to inherit from both
        :class:`~whad.ble.profile.GenericProfile` and :class:`~whad.ble.profile.services.BatteryService`. A
        *generic profile* was then considered as an empty GATT profile that could be used to create default profiles,
        an idea now put aside because it does not fit in our vision of GATT profiles anymore.
    """

__all__ = [
    # Hooks
    "read",
    "write",
    "written",
    "subscribed",
    "unsubscribed",

    # Classes
    "Characteristic",
    "CharacteristicValue",
    "Descriptor",
    "Service",
    "PrimaryService",
    "SecondaryService",
    "ReportReference",
    "UserDescription",
    "ClientCharacteristicConfig",
    "Profile",

    # Old classes (to be removed later)
    "CharacteristicDescriptor",
    "CharacteristicUserDescriptionDescriptor",
    "ReportReferenceDescriptor",
    "GenericProfile",
]



"""Bluetooth Low Energy peripheral utility for WHAD
"""
import json

from whad.cli.app import CommandLineApp, run_app
from whad.ble.cli.peripheral.shell import BlePeriphShell
from whad.ble.utils.clues import CluesDb

from .commands.shell import interactive_handler, check_profile

class BlePeriphApp(CommandLineApp):
    """Bluetooth Low Energy Peripheral emulation app
    """

    def __init__(self):
        """Initialize our application: we need a WHAD interface (adapter) to be
        specified and we support custom commands.
        """
        super().__init__(
            description='WHAD Bluetooth Low Energy peripheral utility',
            commands = True,
            interface = True
        )

        # Add an option to specify the target Bluetooth Device address
        self.add_argument(
            '--bdaddr',
            '-b',
            dest='bdaddr',
            help='Specify target BD address'
        )

        # Add an option to provide an existing device profile
        self.add_argument(
            '--profile',
            '-p',
            dest='profile',
            help='Use a saved device profile'
        )

        # Add an option to allow scripting through a file
        self.add_argument(
            '--file',
            '-f',
            dest='script',
            help='Specify a script to run'
        )

    def run(self):
        """Override App's run() method to handle scripting feature.
        """
        # Launch pre-run tasks
        self.pre_run()

        # Preload CluesDb
        if not CluesDb.load_data():
            self.error("CLUES database could not be loaded, did you use `--recurse-submodules` when cloning the repository ?")

        if self.args.script is not None:
            # If a profile has been provided, load it
            if self.args.profile is not None:
                # Read profile
                with open(self.args.profile,'rb') as f:
                    profile_json = f.read()
                try:
                    if not check_profile(json.loads(profile_json)):
                        self.error("Invalid JSON file (does not contain a valid GATT profile).")
                        profile_json = None
                except json.decoder.JSONDecodeError as parsing_err:
                    self.error((f"Invalid JSON file, parsing error line {parsing_err.lineno}: "
                            f"{parsing_err.msg}"))
                    self.exit()
            else:
                profile_json = None

            # Launch an interactive shell (well, driven by our script)
            myshell = BlePeriphShell(self.interface, profile_json)
            myshell.run_script(self.args.script)
        else:
            # Run the application through WHAD's main app routine
            super().run()

        # Launch post-run tasks
        self.post_run()

def ble_periph_main():
    """Bluetooth Low Energy peripheral emulation main routine.
    """
    app = BlePeriphApp()
    run_app(app)



"""This module provides specific classes to set-up a BLE proxy.

The class :class:`LinkLayerProxy` provides a link-layer level BLE proxy that
will connect to a target device, spawns another device and relay control and
data PDUs between a client and the target device.

The class :class:`GattProxy` provides a GATT BLE proxy that will connect to a
target device, creates another device and relay GATT operations (characteristic
read, write, notification and indication) between a client and the target
device.
"""
import logging

from scapy.layers.bluetooth4LE import BTLE_DATA
from scapy.layers.bluetooth import L2CAP_Hdr

from whad.hub.ble import BDAddress
from whad.ble.connector import Central, Peripheral
from whad.ble.connector.central import CentralDisconnected
from whad.ble.connector.peripheral import PeripheralEventListener, PeripheralEventConnected, \
    PeripheralEventDisconnected
from whad.ble.profile.advdata import AdvDataFieldList, AdvFlagsField, AdvCompleteLocalName
from whad.ble.profile import GenericProfile
from whad.ble.exceptions import HookReturnValue, HookDontForward, ConnectionLostException
from whad.ble.stack.gatt.exceptions import GattTimeoutException
from whad.exceptions import WhadDeviceNotFound
from whad.common.monitors import PcapWriterMonitor, WiresharkMonitor
from whad.hub.ble import Direction as BleDirection


logger = logging.getLogger(__name__)

#############################################
# Link-layer Proxy related classes
#############################################

def is_pdu_valid(pdu):
    """Check if a given PDU is valid (either Data or Control PDU)

    :param Packet pdu: PDU to check
    :rtype: bool
    :return: True if PDU is valid, False otherwise
    """
    if pdu.haslayer(BTLE_DATA):
        btle_data = pdu.getlayer(BTLE_DATA)
        if btle_data.LLID in [0x01, 0x02]:
            return btle_data.haslayer(L2CAP_Hdr)

        # Control PDU
        return btle_data.LLID == 0x03
    return False


def reshape_pdu(pdu):
    """This function remove any SN/NESN/MD bit as it is usually handled by
    the WHAD BLE-compatible dongle. Some BLE controllers and integrated stacks
    do not like to get PDUs with these bits set.

    :param Packet pdu: Bluetooth LE packet to process
    :return Packet: Clean Bluetooth LE packet
    """
    # Reset SN/NESN/MD bits
    btle_data = pdu.getlayer(BTLE_DATA)
    payload = btle_data.payload
    return BTLE_DATA(
        LLID=btle_data.LLID,
        len=len(payload)
    )/payload


class LowLevelPeripheral(Peripheral):
    """Link-layer only Peripheral implementation

    This class is used by the :class:`whad.ble.tools.proxy.LinkLayerProxy` class
    to provide a Link-Layer peripheral with the requested advertising data.
    """

    def __init__(self, proxy, device, adv_data, scan_data, bd_address=None, public=True):
        """Instantiate a LowLevelPeripheral instance

        :param LinkLayerProxy proxy: Reference to the link-layer proxy that will receive events
        :param WhadDevice device: A :class:`whad.device.WhadDevice` object to use
                                  as the physical link
        :param AdvDataFieldList adv_data: Advertising data of the exposed proxy device (mandatory)
        :param AdvDataFieldList scan_data: Scan response data of the exposed proxy
                                           device (optional, can be None)
        """
        super().__init__(device, adv_data=adv_data, scan_data=scan_data, bd_address=bd_address, public=public)
        self.__proxy = proxy
        self.__connected = False
        self.__conn_handle = None
        self.__other_half = None
        self.__pending_data_pdus = []
        self.__pending_control_pdus = []

    @property
    def conn_handle(self):
        return self.__conn_handle

    def set_other_half(self, other_half):
        """Set the *other half*, i.e. the reference to a :class:`LowLevelCentral`
        device connected to the target device

        :param LowLevelCentral other_half: Link-layer Central device to notify events
        """
        self.__other_half = other_half

    def on_connected(self, connection_data):
        """Callback to handle link-layer connection from the underlying peripheral connector

        :param connection_data: Connection data object
        """
        self.__connected = True
        if connection_data.conn_handle is None:
            self.__conn_handle = 0
        else:
            self.__conn_handle = connection_data.conn_handle

        local_peer = BDAddress.from_bytes(
            connection_data.advertiser,
            connection_data.adv_addr_type
        )
        remote_peer = BDAddress.from_bytes(
            connection_data.initiator,
            connection_data.init_addr_type
        )

        # Notify event listener
        self.notify_event(PeripheralEventConnected(
            connection_data.conn_handle,
            local_peer,
            remote_peer
        ))

        # Notify proxy that a connection has been established
        if self.__proxy is not None:
            self.__proxy.on_connect()

        # Foward pending PDUs
        if len(self.__pending_control_pdus) > 0:
            for _pdu in self.__pending_control_pdus:
                if self.__proxy is not None:
                    pdu = self.__proxy.on_ctl_pdu(_pdu, BleDirection.MASTER_TO_SLAVE)
                    if pdu is not None:
                        self.send_pdu(reshape_pdu(pdu), self.__conn_handle)
                else:
                    self.send_pdu(reshape_pdu(_pdu), self.__conn_handle)

        if len(self.__pending_data_pdus) > 0:
            logger.debug("Sending pending PDUs to other half...")
            for _pdu in self.__pending_data_pdus:
                if self.__proxy is not None:
                    logger.debug("Calling proxy on_data_pdu with PDU %s...", _pdu)
                    pdu = self.__proxy.on_data_pdu(_pdu, BleDirection.SLAVE_TO_MASTER)
                    if pdu is not None:
                        self.send_pdu(reshape_pdu(pdu), self.__conn_handle, direction=BleDirection.SLAVE_TO_MASTER)
                else:
                    logger.debug("Directly sending PDU %s...", _pdu)
                    self.send_pdu(reshape_pdu(_pdu), self.__conn_handle)


    def on_ctl_pdu(self, pdu):
        """This method is called whenever a control PDU is received.
        This PDU is then forwarded to the BLE stack to handle it.

        Peripheral devices act as a slave, so we only forward master to slave
        messages to the stack.

        :param Packet pdu: Control PDU received
        """
        if pdu.metadata.direction == BleDirection.MASTER_TO_SLAVE:
            if self.__other_half is not None and self.__connected:
                if self.__proxy is not None:
                    pdu = self.__proxy.on_ctl_pdu(pdu, BleDirection.MASTER_TO_SLAVE)
                    if pdu is not None:
                        self.__other_half.forward_ctrl_pdu(pdu)
                else:
                    self.__other_half.forward_ctrl_pdu(pdu)


    def on_data_pdu(self, pdu):
        """This method is called whenever a data PDU is received.
        This PDU is then forwarded to the BLE stack to handle it.

        :param Packet pdu: Data PDU received
        """
        if pdu.metadata.direction == BleDirection.MASTER_TO_SLAVE:
            if self.__other_half is not None and self.__connected:
                if self.__proxy is not None:
                    pdu = self.__proxy.on_data_pdu(pdu, BleDirection.MASTER_TO_SLAVE)
                    if pdu is not None:
                        self.__other_half.forward_data_pdu(pdu)
                else:
                    self.__other_half.forward_data_pdu(pdu)
            else:
                logger.error("client is not connected to proxy")


    def forward_ctrl_pdu(self, pdu):
        """Forward a control pdu to the target device.

        :param Packet pdu: Control PDU to forward
        """
        logger.info("Forwarding control PDU to central (%s)", str(self.__conn_handle))
        if self.__conn_handle is not None:
            return self.send_pdu(
                reshape_pdu(pdu),
                self.__conn_handle,
                direction=BleDirection.SLAVE_TO_MASTER
            )

        # Not connected, adding control PDU to pending control PDUs
        self.__pending_control_pdus.append(reshape_pdu(pdu))
        return True


    def forward_data_pdu(self, pdu):
        """Forward a data pdu to the target device.

        :param Packet pdu: Data PDU to forward
        """
        if self.__conn_handle is not None:
            logger.info("Forwarding data PDU to central (%s)", str(self.__conn_handle))
            return self.send_pdu(
                reshape_pdu(pdu),
                self.__conn_handle,
                direction=BleDirection.SLAVE_TO_MASTER
            )

        # Not connected, adding data PDU to pending data PDUs
        logger.info("Adding data PDU to pending PDUs")
        self.__pending_data_pdus.append(reshape_pdu(pdu))
        return True


class LowLevelCentral(Central):
    """Link-layer only Central implementation

    This class implements a Central role that is able to initiate a BLE connection
    to a target device and then forward all the control and data PDUs to the attached
    LowLevelPeripheral instance.

    No BLE stack is bound tho this Central role, only raw PDUs sent by the target
    device or received from the associated peripheral.
    """

    def __init__(self, proxy, device, connection_data=None):
        """Instantiate a LowLevelCentral object

        :param WhadDevice device: Underlying WHAD device to use.
        """
        super().__init__(device, existing_connection=connection_data)
        self.__connected = False
        self.__other_half = None
        self.__proxy = proxy


    def set_other_half(self, other_half):
        """Set the *other half*, the associated LowLevelPeripheral.

        :param LowLevelPeripheral other_half: Associated LowLevelPeripheral
        """
        self.__other_half = other_half


    def is_connected(self):
        """Determine if this Central device is connected to the target device.

        :rtype: bool
        :return: True if an active connection exists between the central device
                 and the target, False otherwise
        """
        return self.__connected


    def peripheral(self):
        return True


    def on_connected(self, connection_data):
        """Callback called when our central device is successfully connected to our target device

        :param connection_data: Connection data
        """
        # Call Central.on_connected()
        super().on_connected(connection_data)

        self.__connected = True
        if connection_data.conn_handle is None:
            self.__conn_handle = 0
        else:
            self.__conn_handle = connection_data.conn_handle


    def on_disconnected(self, disconnection_data):
        """Callback called when our central device has been disconnected from our target device
        """
        super().on_disconnected(disconnection_data)

        logger.info("target device has disconnected")
        self.__connected = False
        self.__conn_handle = None


    def on_ctl_pdu(self, pdu):
        """Callback called whenever a Control PDU has been received.

        This callback method then forwards PDU to the associated LowLevelPeripheral object that
        will handle it.

        :param Packet pdu: Received Control PDU
        """
        if pdu.metadata.direction == BleDirection.SLAVE_TO_MASTER:
            if self.__other_half is not None and self.__connected:
                if self.__proxy is not None:
                    pdu = self.__proxy.on_ctl_pdu(pdu, BleDirection.SLAVE_TO_MASTER)
                    if pdu is not None:
                        self.__other_half.forward_ctrl_pdu(pdu)
                else:
                    self.__other_half.forward_ctrl_pdu(pdu)


    def on_data_pdu(self, pdu):
        """Callback called whenever a Data PDU has been received.

        This callback method then forwards PDU to the associated LowLevelPeripheral object that
        will handle it.

        :param Packet pdu: Received Data PDU
        """
        if pdu.metadata.direction == BleDirection.SLAVE_TO_MASTER:
            if self.__other_half is not None and self.__connected:
                if self.__proxy is not None:
                    pdu = self.__proxy.on_data_pdu(pdu, BleDirection.SLAVE_TO_MASTER)
                    if pdu is not None:
                        self.__other_half.forward_data_pdu(pdu)
                else:
                    self.__other_half.forward_data_pdu(pdu)


    def forward_ctrl_pdu(self, pdu):
        """Forward a Control PDU to the connected device, if an active connection exists

        :param Packet pdu: Control PDU to send to the connected device
        """
        if self.__conn_handle is not None:
            return self.send_pdu(reshape_pdu(pdu), self.__conn_handle)

        # Error, proxy not connected
        logger.error("proxy is not connected to target device")
        return False

    def forward_data_pdu(self, pdu):
        """Forward a Data PDU to the connected device, if an active connection exists

        :param Packet pdu: Data PDU to send to the connected device
        """
        if self.__conn_handle is not None:
            return self.send_pdu(reshape_pdu(pdu), self.__conn_handle)

        # Error, proxy not connected
        logger.error("proxy is not connected to target device")
        return False


class LinkLayerProxy:
    """This class implements a GATT proxy that relies on two BLE-compatible
    WHAD devices to create a real BLE device that will proxify all the link-layer
    traffic to another device.
    """

    def __init__(self, proxy=None, target=None, adv_data=None, scan_data=None, bd_address=None,
                 spoof=False, random=False):
        """
        :param BLE proxy: BLE device to use as a peripheral (GATT Server)
        :param BLE target: BLE device to use as a central (GATT Client)
        :param AdvDataFieldList adv_data: Advertising data
        :param AdvDataFieldList scan_data: Scan response data
        :param str bd_address: BD address of target device
        """
        if proxy is None or target is None or bd_address is None:
            raise WhadDeviceNotFound

        if adv_data is None:
            self.__adv_data = AdvDataFieldList(
                AdvFlagsField(),
                AdvCompleteLocalName(b"BleProxy")
            )
        else:
            self.__adv_data = adv_data

        self.__scan_data = scan_data

        # Save both devices
        self.__proxy = proxy
        self.__central = None
        self.__listener = None
        self.__target = target
        self.__peripheral = None
        self.__target_bd_addr = bd_address
        self.__target_random=random
        self.__spoof = spoof


    @property
    def target(self):
        """Proxy target (central)
        """
        return self.__central


    def get_wireshark_monitor(self):
        """Attach a Wireshark monitor to the target device.

        :rtype: WiresharkMonitor
        """
        monitor_ws = WiresharkMonitor()
        monitor_ws.attach(self.__central)
        return monitor_ws


    def get_pcap_monitor(self, filename):
        """Attach a PCAP writer monitor to the target device.

        :rtype: PcapWriterMonitor
        """
        monitor_pcap = PcapWriterMonitor(filename)
        monitor_pcap.attach(self.__central)
        return monitor_pcap


    def close(self):
        """Close proxy
        """
        if self.__central is not None:
            self.__central.close()
        if self.__peripheral is not None:
            self.__peripheral.close()

    def start(self):
        """Start proxy

        The proxy device will be set as a peripheral
        """

        # First, connect our central device to our target device
        logger.info("create low-level central device ...")
        self.__central = LowLevelCentral(self, self.__target)
        self.__central.add_event_handler(self.on_central_event, CentralDisconnected)
        self.__central.lock()
        logger.info("connecting to target device ...")
        if self.__central.connect(self.__target_bd_addr, random=self.__target_random) is not None:
            logger.info("proxy is connected to target device, create our own device ...")

            # Once connected, we start our peripheral
            self.__peripheral = LowLevelPeripheral(
                self,
                self.__proxy,
                self.__adv_data,
                self.__scan_data,
                bd_address=self.__target_bd_addr,
                public=not self.__target_random
            )

            # Interconnect central and peripheral
            logger.info("proxy peripheral device created, interconnect with central ...")
            self.__peripheral.set_other_half(self.__central)
            self.__central.set_other_half(self.__peripheral)
            logger.info("central and peripheral devices are now interconnected")

            # Install peripheral event listener
            self.__listener = PeripheralEventListener(callback=self.on_periph_event)
            self.__peripheral.attach_event_listener(self.__listener)
            self.__listener.start()

            # Start advertising
            logger.info("starting advertising our proxy device")
            self.__peripheral.start()
            self.__central.unlock()
            logger.info("LinkLayerProxy instance is ready")

    def stop(self):
        """Stop proxy
        """
        if self.__listener is not None:
            self.__listener.stop()

        # Stop central
        self.__central.clear_event_handlers()
        self.__central.stop()

    def on_periph_event(self, event):
        """Handle peripheral events.
        """
        if isinstance(event, PeripheralEventConnected):
            """When a central is connected, update its mtu if not 23.
            """
            mtu = self.__central.get_mtu()
            if mtu != 23:
                self.__peripheral.set_mtu(mtu)
        elif isinstance(event, PeripheralEventDisconnected):
            """Notify proxy that peripheral has disconnected
            """
            self.on_disconnect()

    def on_central_event(self, event):
        """Central event handler
        """
        if isinstance(event, CentralDisconnected):
            # Central has disconnected, 
            if self.__peripheral.conn_handle is not None:
                self.__peripheral.disconnect(self.__peripheral.conn_handle)
                self.__peripheral.wait_disconnection()
                self.__peripheral.stop()
                self.__listener.stop()
            else:
                logger.debug("[LinkLayerProxy] peripheral not connected")

            # Reconnect to Central
            self.__central.stop()
            self.start()

    def on_connect(self):
        """This method is called when a client connects to the proxy.
        """
        logger.info("Client has just connected to our proxy")


    def on_disconnect(self):
        """This method is called when a client disconnects from the proxy.
        """
        logger.info("Client has just disconnected from our proxy")


    def on_ctl_pdu(self, pdu, direction):
        """Control PDU callback

        This method is called whenever a BLE control PDU is received. The returned
        PDU will be forwarded to the target device, or discarded if None is returned.

        :param Packet pdu: Scapy packet representing the BLE control PDU
        :return: A PDU to be sent to the target device or None to avoid forwarding.
        :rtype: Packet, None
        """
        logger.info("Received a Control PDU: %s (Direction: %d)", bytes(pdu).hex(), direction)
        return pdu


    def on_data_pdu(self, pdu, direction):
        """Data PDU callback

        This method is called whenever a BLE data PDU is received. The returned
        PDU will be forwarded to the target device, or discarded if None is returned.

        :param Packet pdu: Scapy packet representing the BLE data PDU
        :return: A PDU to be sent to the target device or None to avoid forwarding.
        :rtype: Packet, None
        """
        logger.info("Received a Data PDU: %s (Direction: %d)", bytes(pdu).hex(), direction)
        return pdu


#############################################
# GATT Proxy related classes
#############################################

class ImportedDevice(GenericProfile):
    """Device imported from JSON profile.
    """

    def __init__(self, proxy, target, from_json):
        super().__init__(from_json=from_json)
        self.__target = target
        self.__proxy = proxy


    def on_connect(self, conn_handle):
        """This method is called when a device connects to the GATT proxy, and will
        forward this event to the GATT proxy.
        """
        self.__proxy.on_connect(conn_handle)


    def on_disconnect(self, conn_handle):
        """This method is called when a device disconnects from the GATT proxy and
        forwards this event to the proxy itself.
        """
        self.__proxy.on_disconnect(conn_handle)


    def on_characteristic_read(self, service, characteristic, offset=0, length=0):
        """Callback method that handles a characteristic read operation.

        This method accesses the characteristic value and call a proxy-specific
        callback to let the user choose what should be done with this value.

        If the proxy callback raises a :class:`HookReturnValue` exception, the underlying
        GATT stack will take the provided value and return it to the client.

        If no exception is raised then the original characteristic value is returned.
        """
        try:
            # Get characteristic and read its value
            c = self.__target.get_characteristic(service.uuid, characteristic.uuid)
            value = c.read(offset=offset)

            # Call our proxy characteristic read hook
            self.__proxy.on_characteristic_read(
                service,
                characteristic,
                value,
                offset,
                length
            )

            # By default, return characteristic value
            raise HookReturnValue(value)
        except GattTimeoutException as gatt_error:
            logger.error("GATT timeout during characteristic read, return empty data")
            raise HookReturnValue(b'') from gatt_error
        except ConnectionLostException:
            logger.error("GATT connection lost while reading from characteristic")


    def on_characteristic_write(self, service, characteristic, offset=0, value=b'',
                                without_response=False):
        """Characteristic write hook

        This hook is called whenever a charactertistic is about to be written by a GATT
        client. If this method returns None, then the write operation will return an error.
        """
        c = None

        try:

            # Get target characteristic and write its value
            c = self.__target.get_characteristic(service.uuid, characteristic.uuid)

            self.__proxy.on_characteristic_write(
                service,
                characteristic,
                offset,
                value,
                without_response
            )

            # Write value to target device
            c.write(value, without_response=without_response)
        except HookReturnValue as write_override:
            c.write(write_override.value, without_response=without_response)
        except GattTimeoutException:
            logger.error("GATT timeout during characteristic write")
        except ConnectionLostException:
            logger.error("GATT connection lost while writing to characteristic")


    def on_characteristic_subscribed(self, service, characteristic, notification=False,
                                     indication=False):
        """Characteristic subscription hook.
        """
        try:
            c = self.__target.get_characteristic(service.uuid, characteristic.uuid)
            if notification and c is not None:
                # Forward callback
                def notif_cb(handle, value, indication=False):
                    try:
                        # Forward to proxy
                        self.__proxy.on_notification(
                            service,
                            characteristic,
                            value
                        )

                        # Update characteristic value
                        characteristic.value = value

                    except HookReturnValue as value_override:
                        # Override value if required
                        characteristic.value = value_override.value

                    except HookDontForward:
                        # Don't forward notification
                        pass
                logger.info("[proxy] subscribe to characteristic %s", characteristic.uuid)
                c.subscribe(callback=notif_cb, notification=True)
            elif indication and c is not None:
                # Forward callback
                def indicate_cb(handle, value, indication=True):
                    try:
                        # Forward to proxy hook.
                        self.__proxy.on_indication(
                            service,
                            characteristic,
                            value
                        )

                        # Update characteristic value
                        characteristic.value = value

                    except HookReturnValue as value_override:
                        # Override value if required
                        characteristic.value = value_override.value

                    except HookDontForward:
                        # Don't forward notification
                        pass

                logger.info("[proxy] subscribe to characteristic %s (indication)",
                            characteristic.uuid)
                c.subscribe(callback=indicate_cb, indication=True)

            else:
                logger.error("[proxy] cannot find characteristic %s", characteristic.uuid)

            # No action possible here (for now)
            self.__proxy.on_characteristic_subscribed(
                service,
                characteristic,
                notification=notification,
                indication=indication
            )
        except GattTimeoutException:
            logger.error("GATT timeout during characteristic subscribe")
        except ConnectionLostException:
            logger.error("GATT connection lost while unsubscribing from characteristic")


    def on_characteristic_unsubscribed(self, service, characteristic):
        """Characteristic unsubscription hook.
        """
        try:
            c = self.__target.get_characteristic(service.uuid, characteristic.uuid)
            c.unsubscribe()

            # No action possible here (for now)
            self.__proxy.on_characteristic_unsubscribed(
                service,
                characteristic
            )
        except GattTimeoutException:
            logger.error("GATT timeout during characteristic unsubscribe")
        except ConnectionLostException:
            logger.error("GATT connection lost while unsubscribing from characteristic")

    def on_mtu_changed(self, mtu: int):
        """MTU update callback.
        """
        self.__proxy.on_mtu_changed(mtu)


class GattProxy:
    """GATT Proxy
    """

    def __init__(self, proxy=None, target=None, adv_data=None, scan_data=None, bd_address=None,
                 spoof=False, profile=None, random=False):
        self.__central = None
        self.__peripheral = None
        self.__proxy_dev = proxy
        self.__target_dev = target
        self.__spoof = spoof
        self.__profile = profile
        if adv_data is None:
            self.__adv_data = AdvDataFieldList(
                AdvFlagsField(),
                AdvCompleteLocalName(b"BleProxy")
            )
        else:
            self.__adv_data = adv_data
        self.__scan_data = scan_data
        self.__target = None
        self.__target_bd_addr = bd_address
        self.__target_random = random
        self.__central_mtu = 23
        self.__listener = None

    @property
    def target(self):
        """Proxy target
        """
        return self.__target

    @property
    def central(self):
        """Central role
        """
        return self.__central

    @property
    def peripheral(self):
        """Peripheral role
        """
        return self.__profile

    def get_wireshark_monitor(self):
        """Attach a Wireshark monitor to the our proxy.

        :rtype: WiresharkMonitor
        """
        if self.__peripheral is not None:
            monitor_ws = WiresharkMonitor()
            monitor_ws.attach(self.__peripheral)
            return monitor_ws

        # No wireshark monitor attached
        return None


    def get_pcap_monitor(self, filename):
        """Attach a PCAP writer monitor to the target device.

        :rtype: PcapWriterMonitor
        """
        if self.__peripheral is not None:
            monitor_pcap = PcapWriterMonitor(filename)
            monitor_pcap.attach(self.__peripheral)
            return monitor_pcap

        # No PCAP monitor attached
        return None


    def on_connect(self, conn_handle: int):
        """Connection callback
        """
        logger.info("client connected to proxy (handle: %d)", conn_handle)


    def on_disconnect(self, conn_handle):
        """Disconnection callback
        """
        logger.info("client disconnected from proxy (handle: %d)", conn_handle)


    def on_characteristic_read(self, service, characteristic, value, offset=0, length=0):
        """This callback is called whenever a characteristic is read.

        :param Service service: Service object the target characteristic belongs to.
        :param Characteristic characteristic: Target characteristic object.
        :param bytes value: Characteristic value
        :param int offset: write offset
        :param int length: maximum read length for this characteristic
        """
        logger.info(" << Read characteristic %s: %s", characteristic.uuid, value)


    def on_characteristic_write(self, service, characteristic, offset=0, value=b'',
                                without_response=False):
        """This callback is called whenever a characteristic is written.

        Raise a :class:`HookReturnValue` exception to override the value that will be written
        in the target characteristic.

        Raise any other hook exception (:class:`HookReturnAccessDenied`,
        :class:`HookReturnNotFound`, :class:`HookReturnAuthentRequired` or
        :class:`HookReturnAuthorRequired`) to force the proxy
        to return a specific error to the initiator device.

        :param Service service: Service object the target characteristic belongs to.
        :param Characteristic characteristic: Target characteristic object.
        :param int offset: write offset
        :param bytes value: value to write into this characteristic
        :param bool without_response: True if write operation does not require a response,
                                      False otherwise (default: False)
        """
        logger.info(" >> Write characteristic %s with value %s at offset %d",
                    characteristic.uuid, value, offset)


    def on_characteristic_subscribed(self, service, characteristic, notification=False,
                                     indication=False):
        """This callback is called whenever a characteristic is subscribed to for
        notification or indication.

        :param Service service: Service object the target characteristic belongs to.
        :param Characteristic characteristic: Target characteristic object.
        :param bool notification: Set to True to subscribe for notification
        :param bool indication: Set to True to subscribe to indication
        """
        logger.info(" ** Subscribed to characteristic %s from service %s",
                    characteristic.uuid, service.uuid)


    def on_characteristic_unsubscribed(self, service, characteristic):
        """This callback is called whenever a characteristic is unsubscribed.

        :param Service service: Service object the target characteristic belongs to.
        :param Characteristic characteristic: Target characteristic object.
        """
        logger.info(" ** Unsubscribed to characteristic %s from service %s",
                    characteristic.uuid, service.uuid)


    def on_notification(self, service, characteristic, value):
        """This callback is called whenever a notification is received.
        """
        logger.info(" == notification for characteristic %s from service %s: %s",
                    characteristic.uuid, service.uuid, value)


    def on_indication(self, service, characteristic, value):
        """This callback is called whenever a notification is received.
        """
        logger.info(" == indication for characteristic %s from service %s: %s",
                    characteristic.uuid, service.uuid, value)

    def on_mtu_changed(self, mtu: int):
        """MTU has been updated
        """
        logger.info("== MTU changed to %d", mtu)

    def on_central_event(self, event):
        """Handle Central events.
        """
        if isinstance(event, CentralDisconnected):
            # Central has disconnected, kill peripheral
            if self.__peripheral.conn_handle is not None:
                self.__peripheral.disconnect(self.__peripheral.conn_handle)

            # Reconnect to Central using the already discovered GATT profile
            self.__peripheral.stop()
            self.__central.stop()
            self.__profile = self.__target_profile
            self.start()

    def start(self):
        """Start our GATT Proxy
        """
        logger.info("create our central device")
        self.__central = Central(self.__target_dev, from_json=self.__profile)
        self.__central.add_event_handler(self.on_central_event)
        logger.info("connect to target device (random:%s)", self.__target_random)
        self.__target = self.__central.connect(self.__target_bd_addr, random=self.__target_random)
        if self.__target is not None:
            if self.__profile is not None:
                logger.info("use the provided profile (json) ...")
                self.__target_profile = self.__profile
            else:
                logger.info("connected to target device, discover services and characteristics ...")
                self.__target.discover()
                logger.info("services and characs discovered")
                self.__target_profile = self.__target.export_json()

            # Once connected, we start our peripheral
            self.__central_mtu = self.__central.get_mtu()
            logger.debug("[GattProxy] central mtu: %d", self.__central_mtu)
            logger.info("create a peripheral with similar profile ...")
            self.__profile = ImportedDevice(
                    self,
                    self.__target,
                    self.__target_profile
            )

            # Create a peripheral with the target device address if possible
            # If not, the device will take its own default address.
            self.__peripheral = Peripheral(self.__proxy_dev, profile=self.__profile,
                adv_data=self.__adv_data,
                scan_data=self.__scan_data,
                bd_address=self.__target_bd_addr,
                public=not self.__target_random
            )
            
            # Attach peripheral listener
            self.__listener = PeripheralEventListener(callback=self.on_periph_event)
            self.__peripheral.attach_event_listener(self.__listener)
            self.__listener.start()

            self.__peripheral.enable_peripheral_mode(adv_data=self.__adv_data)

            # Start advertising
            logger.info("starting advertising")
            self.__peripheral.start()
            logger.info("GattProxy instance is ready")

    def stop(self):
        """Stop proxy
        """
        if self.__listener is not None:
            self.__listener.stop()

        # Remove event handlers
        self.central.clear_event_handlers()
        
        # Disconnect central and peripheral
        if self.__target.conn_handle is not None:
            self.__central.disconnect(self.__target.conn_handle)
        self.__central.stop()
        if self.__peripheral is not None:
            if self.__peripheral.conn_handle is not None:
                self.__peripheral.disconnect(self.__peripheral.conn_handle)
            self.__peripheral.stop()

    def on_periph_event(self, event):
        """Handle peripheral events.
        """
        if isinstance(event, PeripheralEventConnected):
            """When a central is connected, update its mtu if not 23.
            """
            self.__central_mtu = self.__central.get_mtu()
            logger.debug("[GattProxy] central mtu: %d", self.__central_mtu)
            if self.__central_mtu != 23:
                logger.debug("[GattProxy] Setting peripheral MTU to %d", self.__central_mtu)
                self.__peripheral.set_mtu(self.__central_mtu)


"""
This module provides some helpers functions and constants related to Bluetooth Low
Energy physical layer.
"""
from enum import IntEnum
from dataclasses import dataclass
from struct import pack

from scapy.layers.bluetooth4LE import BTLE

from whad.helpers import swap_bits
from whad.phy import Endianness, GFSKModulationScheme, PhysicalLayer


class FieldsSize(IntEnum):
    """
    Size of major BLE fields (in bytes).
    """
    ACCESS_ADDRESS_SIZE = 4
    HEADER_SIZE = 2
    CRC_SIZE = 3


def frequency_to_channel(frequency):
    """
    Converts a frequency (in Hz) to the corresponding BLE channel.
    """
    if not isinstance(frequency,int) or frequency < 2402000000 or frequency > 2480000000:
        return None

    frequency = int(frequency / 1000000)

    freq_offset = frequency - 2400
    if freq_offset == 2:
        channel = 37
    elif freq_offset == 26:
        channel = 38
    elif freq_offset == 80:
        channel = 39
    elif freq_offset <= 24:
        channel = int((freq_offset / 2) - 2)
    else:
        channel = int((freq_offset / 2) - 3)

    return channel

def channel_to_frequency(channel):
    """
    Converts a BLE channel to the corresponding frequency (in Hz).
    """
    if not isinstance(channel,int) or channel < 0 or channel > 39:
        return None
    if channel == 37:
        freq_offset = 2
    elif channel == 38:
        freq_offset = 26
    elif channel == 39:
        freq_offset = 80
    elif channel < 11:
        freq_offset = 2 * (channel + 2)
    else:
        freq_offset = 2 * (channel + 3)

    return (2400 + freq_offset) * 1000000

def dewhitening(data, channel):
    """
    Dewhiten data based on BLE channel.
    """
    ret = []
    lfsr = swap_bits(channel) | 2

    for d in data:
        d = swap_bits(d)
        for i in 128, 64, 32, 16, 8, 4, 2, 1:
            if lfsr & 0x80:
                lfsr ^= 0x11
                d ^= i

            lfsr <<= 1
            i >>=1
        ret.append(swap_bits(d))

    return bytes(ret)

def whitening(data, channel):
    """
    Whiten data based on a BLE channel.
    """
    return dewhitening(data, channel)

def crc(data, init=0x555555):
    """
    Computes the 24-bit CRC of provided data.
    """
    ret = [(init >> 16) & 0xff, (init >> 8) & 0xff, init & 0xff]
    for d in data:
        for _ in range(8):
            t = (ret[0] >> 7) & 1

            ret[0] <<= 1
            if ret[1] & 0x80:
                ret[0] |= 1

            ret[1] <<= 1
            if ret[2] & 0x80:
                ret[1] |= 1

            ret[2] <<= 1
            if d & 1 != t:
                ret[2] ^= 0x5b
                ret[1] ^= 0x06

            d >>= 1

    ret[0] = swap_bits(ret[0] & 0xFF)
    ret[1] = swap_bits(ret[1] & 0xFF)
    ret[2] = swap_bits(ret[2] & 0xFF)
    return bytes(ret)



def is_access_address_valid(aa):
    """
    This function checks if the provided access address is valid.
    """
    a = (aa & 0xff000000)>>24
    b = (aa & 0x00ff0000)>>16
    c = (aa & 0x0000ff00)>>8
    d = aa & 0x000000ff
    if a==b and b==c and c==d:
        return False
    if aa == 0x8E89BED6:
        return True
    bb = aa
    for i in range(0,26):
        if (bb & 0x3F) == 0 or (bb & 0x3F) == 0x3F:
            return False
        bb >>= 1
    bb = aa
    t = 0
    a = (bb & 0x80000000)>>31
    for i in range(30,0,-1):
        if (bb & (1<<i)) >> i != a:
            a = (bb & (1<<i))>>i
            t += 1
            if t>24:
                return False
        if (i<26) and (t<2):
            return False
    return True


@dataclass
class BLEPhysicalLayerConfiguration:
    """Bluetooth Low Energy PHY configuration
    """
    address : int = 0x8e89bed6
    channel : int = 37
    crc_init : int = 0x555555


def check_crc(pkt, configuration) -> bool:
    """Check BLE packet CRC depending on the configuration
    """
    return crc(pkt[FieldsSize.ACCESS_ADDRESS_SIZE:-FieldsSize.CRC_SIZE],
               configuration.crc_init) == pkt[-FieldsSize.CRC_SIZE:]

def decoding(pkt, configuration):
    """Decode BLE packet (dewhiten)
    """
    # Remove preamble
    while pkt.startswith(b"\xAA"):
        pkt = pkt[1:]

    dewhitened = dewhitening(pkt[FieldsSize.ACCESS_ADDRESS_SIZE:], configuration.channel)
    size = dewhitened[1]
    packet = pkt[:FieldsSize.ACCESS_ADDRESS_SIZE]
    packet += dewhitened[:size+ FieldsSize.CRC_SIZE + FieldsSize.HEADER_SIZE]
    return packet


PHYS = {
    "LE-1M": PhysicalLayer(
                modulation=GFSKModulationScheme(deviation=250000),
                datarate=1000000,
                endianness=Endianness.LITTLE,
                frequency_range=(2402000000, 2480000000),
                maximum_packet_size=31,
                synchronization_word=b"\xAA",
                configuration=BLEPhysicalLayerConfiguration(),
                scapy_layer=BTLE,
                frequency_to_channel_function=frequency_to_channel,
                channel_to_frequency_function=channel_to_frequency,
                format_address_function=lambda access_address:pack(">I", access_address),
                integrity_function=check_crc,
                encoding_function=None,
                decoding_function=decoding,
    ),

    "LE-2M": PhysicalLayer(
                modulation=GFSKModulationScheme(deviation=500000),
                datarate=2000000,
                endianness=Endianness.LITTLE,
                frequency_range=(2402000000, 2480000000),
                maximum_packet_size=250,
                synchronization_word=b"\xAA\xAA",
                configuration=BLEPhysicalLayerConfiguration(),
                scapy_layer=BTLE,
                frequency_to_channel_function=frequency_to_channel,
                channel_to_frequency_function=channel_to_frequency,
                format_address_function=lambda access_address:pack(">I", access_address),
                integrity_function=crc,
                encoding_function=whitening,
                decoding_function=dewhitening
            )
}



from whad.zigbee.stack.nwk.exceptions import NWKInvalidKey

class NetworkSecurityMaterial:

    KEY_COUNTER = 0

    def __init__(self, key, key_sequence_number=None, outgoing_frame_counter=0):
        if isinstance(key,str):
            try:
                self.key = bytes.fromhex(key.replace(":",""))
            except ValueError:
                raise NWKInvalidKey()
        elif isinstance(key,bytes):
            self.key = key

        if len(self.key) != 16:
            raise NWKInvalidKey()

        if key_sequence_number is not None:
            self.key_sequence_number = key_sequence_number
        else:
            self.key_sequence_number = NetworkSecurityMaterial.KEY_COUNTER
            NetworkSecurityMaterial.KEY_COUNTER+=1

        self.outgoing_frame_counter = outgoing_frame_counter
        self.incoming_frame_counters = {}

    def add_incoming_frame_counter(self, device_address, frame_counter):
        self.incoming_frame_counters[device_address] = frame_counter

    def __eq__(self, other):
        return self.key == other.key

    def __repr__(self):
        printable_key = ":".join(["{:02X}".format(i) for i in self.key])
        return "NetworkSecurityMaterial(Key #{}, '{}')".format(self.key_sequence_number, printable_key)



"""Bluetooth Low Energy emulation tool

This utility will configure a compatible whad device to connect to a given
BLE device, and chain this with another tool.

"""
import json
import logging

from whad.cli.app import CommandLineDevicePipe, run_app
from whad.device import Bridge
from whad.device.connector import LockedConnector
from whad.device.unix import UnixSocketServer, UnixConnector
from whad.hub.ble import Connected, Disconnected, BlePduReceived, BleRawPduReceived, \
    BDAddress
from whad.ble.connector import Peripheral, BLE
from whad.hub.discovery import Capability, Domain

logger = logging.getLogger(__name__)

class BleSpawnOutputPipe(Bridge):
    """wble-spawn output pipe
    """

    def __init__(self, input_connector, output_connector):
        """Custom output pipe, will make sure to forward queued PDUs from
        input_connector to output_connector.
        """
        # Retrieve an instance of our protocol hub
        self.hub = input_connector.hub

        # Do we get raw PDUs ?
        if input_connector.support_raw_pdu():
            self.support_raw_pdu = True
        else:
            self.support_raw_pdu = False

        # Save input connector connection handle, as we are supposed to be
        # connected.
        self.__in_handle = input_connector.conn_handle

        # This will dissociate the underlying devices from our connectors.
        # Our central proxy will no more store queued packets, so we are left
        # with the ones we already captured.
        super().__init__(input_connector, output_connector)


    def convert_packet_message(self, message, ingress=True):
        """Convert a BleRawPduReceived/BlePduReceived notification into the
        corresponding SendBleRawPdu/SendBlePdu command, using the provided
        connection handle.
        """
        if ingress:
            connector = self.input
        else:
            connector = self.output

        # Do we received a packet notification ?
        logger.debug("[ble-spawn][output-pipe] convert message %s into a command", message)
        if isinstance(message, BleRawPduReceived):
            # Does our input connector support raw packets ?
            if connector.support_raw_pdu():
                logger.debug("[ble-spawn][output-pipe] connector supports raw pdu")
                # Create a SendBleRawPdu command
                command = connector.hub.ble.create_send_raw_pdu(
                    message.direction,
                    message.pdu,
                    message.crc,
                    encrypt=False,
                    access_address=message.access_address,
                    conn_handle=message.conn_handle, # overwrite the connection handle
                )
                logger.debug("[ble-spawn][output-pipe] created command %s", command)
            else:
                logger.debug("[ble-spawn][output-pipe] connector does not support raw pdu")
                # Create a SendBlePdu command
                command = connector.hub.ble.create_send_pdu(
                    message.direction,
                    message.pdu,
                    self.__in_handle, # overwrite the connection handle
                    encrypt=False
                )
                logger.debug("[ble-spawn][output-pipe] created command %s", command)
        elif isinstance(message, BlePduReceived):
            # Does our input connector support raw packets ?
            if connector.support_raw_pdu():
                logger.debug("[ble-spawn][output-pipe] connector supports raw pdu")
                # Create a SendBleRawPdu command
                command = connector.hub.ble.create_send_raw_pdu(
                    message.direction,
                    message.pdu,
                    None,
                    encrypt=False,
                    access_address=0x11223344, # We use the default access address
                    conn_handle=self.__in_handle, # overwrite the connection handle
                )
                logger.debug("[ble-spawn][output-pipe] created command %s", command)
            else:
                logger.debug("[ble-spawn][output-pipe] connector does not support raw pdu")
                # Create a SendBlePdu command
                command = self.input.hub.ble.create_send_pdu(
                    message.direction,
                    message.pdu,
                    self.__in_handle, # overwrite the connection handle
                    encrypt=False
                )
                logger.debug("[ble-spawn][output-pipe] created command %s", command)
        else:
            # Not a BLE packet notification
            command = None

        # Return generated command
        return command

    def on_inbound(self, message):
        """Process inbound messages.

        Inbound packets are packets coming from our output connector, i.e. the
        central device connected to our emulated peripheral, that need to be
        forwarded as packets to the previous tool.

        Normally, since we get packets from a central device we are supposed to
        be connected and know the connection handle corresponding to this
        connection.
        """
        logger.info("spawn inbound: %s", message)
        if isinstance(message, (BleRawPduReceived, BlePduReceived)):
            logger.debug(
                "[ble-spawn][output-pipe] received an inbound PDU message %s",
                message
            )
            command = self.convert_packet_message(message, True)
            self.input.send_command(command)
        else:
            super().on_inbound(message)

    def on_outbound(self, message):
        logger.info("spawn outbound: %s", message)

        # Update connection handle if a new connection is created
        if isinstance(message, Connected):
            logger.debug(
                "[wble-spawn][output-pipe] new connection, update connection handle to %d",
                message.conn_handle
            )
            self.__in_handle = message.conn_handle

            # If input interface has been locked, unlock it.
            if self.input.is_locked():
                self.input.unlock()
        elif isinstance(message, Disconnected):
            logger.debug(
                "[wble-spawn][output-pipe] Central has just disconnected."
            )
            # We are now disconnected, we must lock our input interface
            self.input.lock()

        # Forward the message to the bridged interface
        super().on_outbound(message)

class BleSpawnInputPipe(Bridge):
    """ble-spawn input pipe

    When ble-spawn is chained after another whad tool, it spawns a device
    based on the specified profile using the specified WHAD adapter, awaits for
    a connection and then send every receive packets to the previous tool as
    well as relay packets it receives to the connected central device.
    """

    def __init__(self, input_connector, output_connector):
        """Initialize our ble-spawn input pipe.
        """
        self.__output_pending_packets = []

        # Do we get raw PDUs ?
        if output_connector.support_raw_pdu():
            self.support_raw_pdu = True
        else:
            self.support_raw_pdu = False

        logger.debug('[ble-spawn][input-pipe] Initialization')
        super().__init__(input_connector, output_connector)

        logger.debug('[ble-spawn][input-pipe] Initialize properties')
        self.__connected = False
        self.__in_conn_handle = None
        self.__out_conn_handle = None

    def set_in_conn_handle(self, conn_handle: int):
        """Saves the input connector connection handle.
        """
        self.__in_conn_handle = conn_handle

    def set_out_conn_handle(self, conn_handle: int):
        """Saves output connection handle.
        """
        logger.debug("[ble-spawn][input-pipe] set output connection handle to %d", conn_handle)
        self.__out_conn_handle = conn_handle

    def convert_packet_message(self, message, conn_handle, ingress=True):
        """Convert a BleRawPduReceived/BlePduReceived notification into the
        corresponding SendBleRawPdu/SendBlePdu command, using the provided
        connection handle.
        """
        if ingress:
            connector = self.input
        else:
            connector = self.output

        # Do we received a packet notification ?
        logger.debug("[ble-spawn][input-pipe] convert message %s into a command", message)
        if isinstance(message, BleRawPduReceived):
            # Does our input connector support raw packets ?
            logger.debug("[ble-spawn][input-pipe] connector does not support raw pdu")
            # Create a SendBlePdu command
            command = connector.hub.ble.create_send_pdu(
                message.direction,
                message.pdu,
                conn_handle, # overwrite the connection handle
                encrypt=False
            )
            logger.debug("[ble-spawn][input-pipe] created command %s", command)
        elif isinstance(message, BlePduReceived):
            # Does our input connector support raw packets ?
            logger.debug("[ble-spawn][input-pipe] connector does not support raw pdu")
            # Create a SendBlePdu command
            command = self.input.hub.ble.create_send_pdu(
                message.direction,
                message.pdu,
                conn_handle, # overwrite the connection handle
                encrypt=False
            )
            logger.debug("[ble-spawn][input-pipe] created command %s", command)
        else:
            # Not a BLE packet notification
            command = None

        # Return generated command
        return command

    def on_inbound(self, message):
        """Process inbound messages.

        Inbound packets are packets coming from our output connector, i.e. the
        central device connected to our emulated peripheral, that need to be
        forwarded as packets to the previous tool.

        Normally, since we get packets from a central device we are supposed to
        be connected and know the connection handle corresponding to this
        connection.
        """
        if isinstance(message, (BleRawPduReceived, BlePduReceived)):
            if not self.__connected:
                logger.debug(
                    "[ble-spawn][input-pipe] add pending inbound PDU message %s to queue",
                    message
                )
            else:
                logger.debug(
                    "[ble-spawn][input-pipe] received an inbound PDU message %s",
                    message
                )
                command = self.convert_packet_message(message, self.__in_conn_handle, True)
                self.input.send_command(command)
        elif isinstance(message, Disconnected):
            # Central device has disconnected, we don't care but we don't send this
            # notification to our chained tool.
            logger.debug('[ble-spawn][input-pipe] received a disconnection notification, discard')
            return
        elif isinstance(message, Connected):
            logger.debug((
                "[ble-spawn][input-pipe] received a connection notification, "
                "update input conn_handle to %d"),
                message.conn_handle
            )
            # Central device has connected, update our output connection handle.
            self.set_out_conn_handle(message.conn_handle)
            self.__connected = True

            # Send pending packets, if any
            for message in self.__output_pending_packets:
                logger.debug('[ble-spawn][input-pipe] process pending PDU message %s', message)
                command = self.convert_packet_message(message, self.__out_conn_handle, False)
                self.output.send_command(command)
        else:
            logger.debug("[ble-spawn][input-pipe] forward default inbound message %s", message)
            # Forward other messages
            super().on_inbound(message)

    def on_outbound(self, message):
        """Process outbund messages.
        """
        if isinstance(message, (BleRawPduReceived, BlePduReceived)):
            if self.__out_conn_handle is not None:
                logger.debug("[ble-spawn][input-pipe] received an outbound PDU message %s", message)
                command = self.convert_packet_message(message, self.__out_conn_handle, False)
                self.output.send_command(command)
            else:
                logger.debug("[ble-spawn][input-pipe] not connected but received %s", message)
                # Save packet as message in pending packets
                self.__output_pending_packets.append(message)
        elif isinstance(message, Connected):
            # Don't forward this message.
            logger.debug("[ble-spawn][input-pipe] received a connection notification, discard")
            return
        elif isinstance(message, Disconnected):
            # Chained tool has lost connection, we must handle it
            logger.debug("[ble-spawn][input-pipe] received a disconnection notification, discard")
            return
        else:
            logger.debug("[ble-spawn][input-pipe] forward default outbound message %s", message)
            # Forward other messages
            super().on_outbound(message)


class BleSpawnApp(CommandLineDevicePipe):
    """Bluetooth Low Energy device spawning tool.
    """

    MODE_END_CHAIN = 0
    MODE_START_CHAIN = 1

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD Bluetooth Low Energy device emulation tool',
            interface=True,
            commands=False
        )

        self.add_argument(
            '--profile',
            '-p',
            dest='profile',
            help='Use a saved device profile'
        )

        self.input_conn_handle = None
        self.output_conn_handle = None

    def run(self, pre: bool = True, post: bool = True):
        """Override App's run() method to handle scripting feature.
        """
        try:
            # Launch pre-run tasks
            if pre:
                self.pre_run()

            # We need to have an interface specified
            if self.interface is not None:

                if self.args.profile is not None:
                    try:
                        # Load file content
                        with open(self.args.profile,'rb') as profile:
                            profile_json = profile.read()
                            profile = json.loads(profile_json)

                        adv_data = bytes.fromhex(profile["devinfo"]["adv_data"])
                        scan_rsp = bytes.fromhex(profile["devinfo"]["scan_rsp"])

                        # If stdin is piped, we are supposed to advertise a device and
                        # proxify once connected
                        if self.is_stdin_piped() and not self.is_stdout_piped():
                            # We create a peripheral that will send all packets to our input interface
                            self.create_input_proxy(adv_data, scan_rsp, int(self.args.conn_handle),
                                                    profile)

                        # Else if stdout is piped, we are supposed to advertise a device
                        # and proxify when connected
                        elif self.is_stdout_piped() and not self.is_stdin_piped():
                            # We create a peripheral that will proxy all messages
                            self.create_output_proxy(adv_data, scan_rsp)
                        else:
                            self.error('Tool must be piped to another WHAD tool.')
                    except FileNotFoundError:
                        self.error(f"Profile file not found ({self.args.profile}).")
                    except IOError:
                        self.error(f"Cannot read file ({self.args.profile}).")
                else:
                    self.error("You need to specify a profile file with option --profile.")
            else:
                self.error("You need to specify an interface with option --interface.")

        except KeyboardInterrupt:
            self.warning("ble-spawn stopped (CTL-C)")

        # Launch post-run tasks
        if post:
            self.post_run()

    def create_input_proxy(self, adv_data: bytes, scan_data: bytes, conn_handle, profile):
        """Configure our hardware to advertise a BLE peripheral, and once
        a central device is connected relay all packets to our input_interface.
        """
        self.input_conn_handle = int(self.args.conn_handle)

        # Extract address info from profile
        address = profile["devinfo"]["bd_addr"]
        address_type = profile["devinfo"]["addr_type"]

        # Create our peripheral
        logger.info("[ble-spawn] Creating peripheral ...")
        peripheral = Peripheral(self.interface, adv_data=adv_data, scan_data=scan_data,
                                bd_address=address, public=address_type==BDAddress.PUBLIC)

        # Query the input device to check if it supports raw BLE packets
        self.input_interface.discover()
        capabilities = self.input_interface.get_domain_capability(Domain.BtLE)
        can_send_raw = not (capabilities & Capability.NoRawData)
        if can_send_raw and not peripheral.support_raw_pdu():
            self.warning((
                "WHAD interface used as fake BLE device does not support raw PDUs, "
                "while the other end does.\n    Some control PDUs may be lost in"
                "this packet processing chain and may cause issues !"
            ))

        # Query the input device to check if it supports raw BLE packets
        self.input_interface.discover()
        capabilities = self.input_interface.get_domain_capability(Domain.BtLE)
        can_send_raw = not (capabilities & Capability.NoRawData)
        if can_send_raw and not peripheral.support_raw_pdu():
            self.warning((
                "WHAD interface used as fake BLE device does not support raw PDUs, "
                "while the other end does.\n    Some control PDUs may be lost in"
                "this packet processing chain and may cause issues !"
            ))

        # Create our packet bridge
        logger.info("[ble-spawn] Starting our input pipe")
        in_pipe = BleSpawnInputPipe(LockedConnector(self.input_interface), peripheral)
        in_pipe.set_in_conn_handle(conn_handle)

        # Loop until the user hits CTL-C
        in_pipe.join()

    def create_output_proxy(self, adv_data, scan_data):
        """Create an output proxy that will relay packets from our emulated BLE
        peripheral to a chained tool.
        """
        # Create our peripheral
        logger.info("[ble-spawn] Creating peripheral ...")
        peripheral = Peripheral(self.interface, adv_data=adv_data, scan_data=scan_data)
        peripheral.lock()
        peripheral.wait_connection()

        # Create our unix socket server
        unix_server = UnixConnector(UnixSocketServer(parameters={
            'domain': 'ble'
        }))

        # Create our packet bridge
        logger.info("[ble-spawn] Starting our output pipe")
        out_pipe = BleSpawnOutputPipe(peripheral, unix_server)

        # Loop until the user hits CTL-C
        out_pipe.join()

        logger.warning("Unix socket client disconnected")

def ble_spawn_main():
    """BLE device spawning tool main routine.
    """
    app = BleSpawnApp()
    run_app(app)



"""This module provides a UartIface class that can be used with a WhadDeviceConnector
to interact with a WHAD-enable firmware device that uses UART as its transport layer.

This class handles device connection, disconnection and read/write operations. All the
parsing magic is performed in our WhadDevice class.
"""

import os
import select
from time import sleep
from queue import Empty

# Import serial
from serial import Serial
from serial.tools.list_ports import comports

from whad.exceptions import WhadDeviceNotReady, WhadDeviceError
from whad.helpers import message_filter
from whad.exceptions import WhadDeviceNotFound

from whad.hub.generic.cmdresult import CommandResult, Success
from whad.hub.discovery import DeviceReady

from .device import Device

SUPPORTED_UART_DEVICES = (
    (0xc0ff, 0xeeee, "WHAD", "ButteRFly dongle"), # Butterfly Dongle
    (0x303A, None, None, None),   # Espressif ESP-32 board
    (0x10C4, 0xEA60, None, None), # Espressif ESP-32 CP2102 board
    (0x0483, 0x374e, None, None), # Nucleo WL55
)

def get_port_info(port):
    """Find information about a serial port

    :param string port: Target serial port
    """
    for p in comports():
        if p.device == port:
            return p
    return None

def is_device_supported(vid, pid, manufacturer, product):
    """Check if a device is supported by WHAD.
    """
    for devinfo in SUPPORTED_UART_DEVICES:
        _vid, _pid, _manuf, _product = devinfo
        if _vid is not None and _vid != vid:
            continue
        if _pid is not None and _pid != pid:
            continue
        if _manuf is not None and _manuf != manufacturer:
            continue
        if _product is not None and _product != product:
            continue

        # Device is supported.
        return True

    # Device is not supported.
    return False

class Uart(Device):
    """
    UartIface device class.
    """
    INTERFACE_NAME = "uart"

    @classmethod
    def list(cls):
        '''
        Returns a list of available UART devices.

        To prevent identifying serial ports which are not compatible with WHAD, it implements
        a filtering mechanism based on vid, pid, manufacturer and / or product.
        '''

        devices = []
        for uart_dev in comports():
            if is_device_supported(uart_dev.vid, uart_dev.pid, uart_dev.manufacturer,
                                   uart_dev.product):
                dev = Uart(uart_dev.device, baudrate=115200)
                devices.append(dev)
        return devices


    def __init__(self, port='/dev/ttyUSB0', baudrate=115200):
        """
        Create device connection
        """
        super().__init__()

        # Connect to target UART device in non-blocking mode
        self.__port = port
        self.__baudrate = baudrate
        self.__fileno = None
        self.__uart = None
        self.__opened = False

        # Determine if device is CDC ACM (usb subsystem)
        port_info = get_port_info(self.__port)
        if port_info is None:
            raise WhadDeviceNotFound

        # 'subsystem' is only defined on Linux serial ports, we check if
        # this property is missing to handle Mac OS (Windows system is
        # marked as unsupported for now)
        if not hasattr(port_info, "subsystem"):
            # Determine if device is ACM on Mac OS
            self.__is_acm = port_info.usb_info is not None
        else:
            # On Linux systems, ACM devices use the "usb" subsystem.
            self.__is_acm = port_info.subsystem == "usb"

    @property
    def identifier(self):
        '''
        Returns the identifier of the device (e.g., serial port).
        '''
        return self.__port

    def is_acm(self):
        """Determine if this UART device is a CDC ACM one.

        :return: True if CDC ACM, False otherwise.
        """
        return self.__is_acm


    def open(self):
        """
        Open device.
        """
        if not self.__opened:
            # Open UART device
            self.__uart = Serial(self.__port, self.__baudrate)

            # Get file number to use with select()
            self.__fileno = self.__uart.fileno()
            self.__opened = True

            # If device is CDC ACM, we don't need to reset it
            # only reset true serial devices
            if not self.__is_acm:
                self.__uart.dtr = False             # Non reset state
                self.__uart.rts = False             # Non reset state
                self.__uart.dtr = self.__uart.dtr   # usbser.sys workaround

            # Ask parent class to run a background I/O thread
            super().open()


    def reset(self):
        """Reset device.

        This routine tries to reset device by setting RTS to high. This works
        on NodeMCU chips as well as any Arduino-compatible device.
        """
        # If device is a true serial device, ask for a reset through DTR/RTS
        if not self.__is_acm:
            # Reset device through DTR
            self.__uart.dtr = False             # Non reset state
            self.__uart.rts = True             # Non reset state
            sleep(0.2)
            self.__uart.dtr = False             # Non reset state
            self.__uart.rts = False             # Non reset state

            try:
                # Wait for a ready message for 1 second
                msg = self.wait_for_single_message(
                    1.0,
                    message_filter(DeviceReady)
                )
                self.put_message(msg)
            except Empty:
                # Use the classic way to reset device (RTS-based reset failed)
                msg = self.hub.discovery.create_reset_query()
                self.send_command(
                    msg,
                    message_filter(DeviceReady)
                )
        else:
            # Device is ACM, send a classic reset message to device
            msg = self.hub.discovery.create_reset_query()
            self.send_command(
                msg,
                message_filter(DeviceReady)
            )


    def close(self):
        """
        Close current device.
        """
        # Ask parent class to stop I/O thread
        super().close()

        # Close underlying device.
        if self.__uart is not None:
            self.__uart.close()
        self.__uart = None
        self.__fileno = None
        self.__opened = False

    def write(self, payload):
        """Writes payload to the device. It relies on select() in order to make sure
        we are allowed to write to the device and wait without eating too much CPU
        if the device is not ready to be written to.

        :param bytes payload: Data to write
        :returns: number of bytes written to the device
        """
        try:
            if not self.__opened:
                raise WhadDeviceNotReady()

            nb_bytes_written = 0
            wlist = [self.__fileno]
            elist = [self.__fileno]
            _, writers, __ = select.select(
                [],
                wlist,
                elist
            )

            if len(writers) > 0:
                nb_bytes_written = os.write(self.__fileno, payload)
            return nb_bytes_written
        except OSError as os_error:
            raise WhadDeviceError("Sending data to WHAD device failed.") from os_error

    def read(self):
        """Fetches data from the device, if there is any data to read. We call select()
        to make sure data is waiting to be read before reading it. Data is then sent to
        our parsing method through on_data_received() that will handle data reassembling
        and message parsing and dispatch.
        """

        if not self.__opened:
            raise WhadDeviceNotReady()

        rlist = [self.__fileno]
        wlist = []
        elist = [self.__fileno]

        readers, _, __ = select.select(
            rlist,
            wlist,
            elist,
            1
        )

        if not self.__opened:
            raise WhadDeviceNotReady()

        # Handle incoming messages if any
        if len(readers) > 0 and self.__fileno is not None:
            data = os.read(self.__fileno, 1024)

            # Make sure data is not empty (detect device disconnection/malfunction)
            if len(data) == 0 :
                # Device does not behave as expected, may not be ready.
                raise WhadDeviceNotReady()

            # Feed our IO thread with received data
            return data

        # Nothing to read
        return None

    def change_transport_speed(self, speed):
        """Set UART speed for true serial devices. CDC ACM devices will ignore
        it.

        :param int speed: New baudrate to apply to current serial device.
        """
        if not self.__is_acm:
            #msg = Message()
            #msg.discovery.set_speed.speed = speed
            msg = self.hub.discovery.create_set_speed(speed)
            resp = self.send_command(
                msg,
                message_filter(CommandResult)
            )

            if isinstance(resp, Success):
                # Change baudrate
                self.__uart.baudrate = speed

                # Wait for device to be ready (500ms)
                sleep(0.5)



from whad.dot15d4.stack.mac.constants import MACAddressMode
from whad.dot15d4.stack.manager import Dot15d4Manager
from whad.zigbee.stack.apl.zdo import ZigbeeDeviceObjects
from whad.common.stack import alias, source, state # state may not be necessary ?
import logging

logger = logging.getLogger(__name__)


@alias('apl')
class APLManager(Dot15d4Manager):
    """
    This class implements the Zigbee Application layer manager (APL).
    It exposes Zigbee Device Object and Application Object.

    Note: this component is a bit particular in ZigBee stack since it
    implements Applications instead of Services. The design adopted here
    keeps the Manager as Layer / Entry point and override Services by Applications.
    We also introduce a dictionnary mapping endpoints to applications and
    methods facilitating applications selection & management.
    """

    def init(self):
        """
        Initialize the APL layer.
        """
        # Initialize Endpoints
        self.endpoints = {}

        self.attach_application(ZigbeeDeviceObjects(), 0)

    # API to manipulate applications
    def initialize(self):
        """
        This method initializes all applications.
        """
        for app in self.endpoints.values():
            app.initialize()

    def start(self):
        """
        This method starts all applications.
        """
        for app in self.endpoints.values():
            app.start()

    def attach_application(self, app, endpoint):
        """
        This method attaches a new application to the layer on a specific endpoint.
        """
        # Provide a reference to the manager in application.
        app.manager = self

        # Add app as service
        self._services[app.name] = app
        # Add app in the endpoints dictionnary
        self.endpoints[endpoint] = app

    def get_endpoints(self):
        """
        This method returns a list of all the local endpoints.
        """
        return list(self.endpoints.keys())

    def get_application_by_endpoint(self, endpoint):
        """
        This method returns the application linked to the provided endpoint (if any).
        """
        if endpoint in self.endpoints:
            return self.endpoints[endpoint]
        return None

    def get_application_by_name(self, name):
        """
        This method returns the application linked to the provided name (if any).
        """
        if name in self._services:
            return self._services[name]
        return None

    def get_applications(self):
        """
        Iterator over all registered applications.
        """
        for application in self.endpoints.values():
            yield application


    def send_interpan_data(
                            self,
                            asdu,
                            asdu_handle=0,
                            source_address_mode=MACAddressMode.SHORT,
                            destination_pan_id=0xFFFF,
                            destination_address=0xFFFF,
                            destination_address_mode=MACAddressMode.SHORT,
                            profile_id=0,
                            cluster_id=0,
                            acknowledged_transmission=False
    ):
        """
        Transmits InterPAN PDU to the APS layer.
        """
        return self.get_layer('aps').get_service("interpan").interpan_data(
            asdu,
            asdu_handle=asdu_handle,
            source_address_mode=source_address_mode,
            destination_pan_id=destination_pan_id,
            destination_address=destination_address,
            destination_address_mode=destination_address_mode,
            profile_id=profile_id,
            cluster_id=cluster_id,
            acknowledged_transmission=acknowledged_transmission
        )

    def send_data(
                    self,
                    asdu,
                    destination_address_mode,
                    destination_address,
                    destination_endpoint,
                    alias_address=None,
                    alias_sequence_number=0,
                    radius=30,
                    security_enabled_transmission=False,
                    use_network_key=False,
                    acknowledged_transmission=False,
                    fragmentation_permitted=False,
                    include_extended_nonce=False,
                    cluster_id=None,
                    profile_id=None,
                    application=None
    ):
        """
        Transmits Data PDU to the APS layer.
        """
        selected_endpoint = None
        for endpoint, running_application in self.endpoints.items():
            if application == running_application:
                selected_endpoint = endpoint
                break
        if selected_endpoint is None:
            return False

        return self.get_layer('aps').get_service("data").data(
            asdu,
            destination_address_mode,
            destination_address,
            destination_endpoint,
            profile_id,
            cluster_id,
            selected_endpoint,
            alias_address=alias_address,
            alias_sequence_number=alias_sequence_number,
            radius=radius,
            security_enabled_transmission=security_enabled_transmission,
            use_network_key=use_network_key,
            acknowledged_transmission=acknowledged_transmission,
            fragmentation_permitted=fragmentation_permitted,
            include_extended_nonce=include_extended_nonce
        )

    # Callbacks processing the informations transmitted by the APS layer
    @source('aps', "INTRP-DATA")
    def on_intrp_data(
                        self,
                        asdu,
                        profile_id=0,
                        cluster_id=0,
                        destination_pan_id=0xFFFF,
                        destination_address=0xFFFF,
                        source_pan_id=0xFFFF,
                        source_address=0xFFFF,
                        link_quality=255
    ):
        """
        Callback processing InterPAN PDU transmitted by the APS layer.
        """
        # Iterate over registered applications
        for application in self.endpoints.values():
            # If we found an application matching the profile id, forwards to the cluster
            if application.profile_id == profile_id:
                application.on_interpan_data(
                    asdu,
                    cluster_id=cluster_id,
                    destination_pan_id=destination_pan_id,
                    destination_address=destination_address,
                    source_pan_id=source_pan_id,
                    source_address=source_address,
                    link_quality=link_quality
                )

    @source('aps', "APSDE-DATA")
    def on_apsde_data(
                        self,
                        asdu,
                        destination_address,
                        destination_address_mode,
                        destination_endpoint,
                        source_address,
                        source_address_mode,
                        source_endpoint,
                        profile_id,
                        cluster_id,
                        security_status,
                        link_quality
    ):
        """
        Callback processing data PDU transmitted by the APS layer.
        """

        # Checks if an endpoint matches
        if destination_endpoint not in self.endpoints:
            logger.info("[apl] destination endpoint not found.")
            return

        application = self.endpoints[destination_endpoint]

        # Checks if the profile id matches the one used by the application
        if application.profile_id != profile_id:
            logger.info("[apl] profile identifier doesn't match the application profile.")
            return

        # If we got a match, forward to application data callback
        success = application.on_data(
            asdu,
            source_address,
            source_address_mode,
            cluster_id,
            security_status,
            link_quality
        )
        # If cluster has not been found in the corresponding application, display an error
        if not success:
            logger.info("[apl] cluster not found (cluster_id=0x%04x).", cluster_id)

    @source('aps', "APSME-JOIN")
    def on_join(
                                self,
                                network_address,
                                extended_address,
                                capability_information,
                                rejoin=False,
                                secure_rejoin=False
    ):
        self.get_service("zdo").on_join(
            network_address,
            extended_address,
            capability_information,
            rejoin=False,
            secure_rejoin=False
        )

    @source('aps', "APSME-TRANSPORT-KEY")
    def on_apsme_transport_key(
                                self,
                                transport_key_data,
                                source_address,
                                standard_key_type
    ):
        """
        Callback processing transport key indicated by the APS layer.
        """
        # Forward to ZDO application
        self.get_service("zdo").on_transport_key(
            source_address,
            standard_key_type,
            transport_key_data
        )



"""WHAD Protocol Dot15d4 address messages abstraction layer.
"""
from whad.protocol.whad_pb2 import Message
from whad.protocol.dot15d4.dot15d4_pb2 import StartCmd, StopCmd
from whad.hub.events import JammedEvt
from ..message import pb_bind, PbFieldInt, PbMessageWrapper
from . import Dot15d4Domain

@pb_bind(Dot15d4Domain, 'sniff', 1)
class SniffMode(PbMessageWrapper):
    """Dot15d4 sniffing mode
    """

    channel = PbFieldInt('dot15d4.sniff.channel')

@pb_bind(Dot15d4Domain, 'jam', 1)
class JamMode(PbMessageWrapper):
    """Dot15d4 jamming mode
    """

    channel = PbFieldInt('dot15d4.jam.channel')

@pb_bind(Dot15d4Domain, 'end_device', 1)
class EndDeviceMode(PbMessageWrapper):
    """Dot15d4 end device mode
    """

    channel = PbFieldInt('dot15d4.end_device.channel')

@pb_bind(Dot15d4Domain, 'router', 1)
class RouterMode(PbMessageWrapper):
    """Dot15d4 router mode
    """

    channel = PbFieldInt('dot15d4.router.channel')

@pb_bind(Dot15d4Domain, 'coordinator', 1)
class CoordMode(PbMessageWrapper):
    """Dot15d4 coordinator mode
    """

    channel = PbFieldInt('dot15d4.coordinator.channel')

@pb_bind(Dot15d4Domain, 'ed', 1)
class EnergyDetectionMode(PbMessageWrapper):
    """Dot15d4 energy detection mode
    """

    channel = PbFieldInt('dot15d4.ed.channel')

@pb_bind(Dot15d4Domain, 'mitm', 1)
class MitmMode(PbMessageWrapper):
    """Dot15d4 man-in-the-middle mode
    """

    role = PbFieldInt('dot15d4.mitm.role')

@pb_bind(Dot15d4Domain, 'start', 1)
class Start(PbMessageWrapper):
    """Dot15d4 start mode message class
    """

    def __init__(self, message: Message = None):
        super().__init__(message=message)
        self.message.dot15d4.start.CopyFrom(StartCmd())

@pb_bind(Dot15d4Domain, 'stop', 1)
class Stop(PbMessageWrapper):
    """Dot15d4 stop mode message class
    """

    def __init__(self, message: Message = None):
        super().__init__(message=message)
        self.message.dot15d4.stop.CopyFrom(StopCmd())

@pb_bind(Dot15d4Domain, 'jammed', 1)
class Jammed(PbMessageWrapper):
    """Dot15d4 Jammed notification
    """

    timestamp = PbFieldInt('dot15d4.jammed.timestamp')

    def to_event(self) -> JammedEvt:
        """Convert this message into a WHAD event.
        """
        return JammedEvt(
            timestamp=self.timestamp
        )
    
    @staticmethod
    def from_event(event):
        """Convert an event into a message.
        """
        return Jammed(
            timestamp=event.timestamp
        )

@pb_bind(Dot15d4Domain, 'ed_sample', 1)
class EnergyDetectionSample(PbMessageWrapper):
    """Dot15d4 Jammed notification
    """

    sample = PbFieldInt('dot15d4.ed_sample.sample')
    timestamp = PbFieldInt('dot15d4.ed_sample.timestamp')
